/* As of v.0.52.0, this is broken (incompatible with changes to GGNFS).
   Don't use it! If you really want to, fix it to reflect the changes
   in blanczos32.c
*/
/**************************************************************/
/* blanczos64.c                                               */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "prand.h"


#define BIT(_i) ((0x0000000000000001ULL)<<(_i))

/* I didn't want to include ggnfs.h for this one lousy prototype.
*/
void printTmp(char *fmt, ...);


/*************************************************************/ 
/* Implementation of the Block Lanczos algorithm for finding */
/* vectors in the kernel of a large sparse matrix.           */
/* The implementation here is as descibed in:                */
/* ``A Block Lanczos Algorithm for Finding Dependencies over */
/*   GF(2)'', Peter Montgomery.                              */
/* The paper I have does not have a publication source on it,*/
/* so it's probably a preprint.                              */
/*************************************************************/ 
/* This version uses the sparse matrix as a black box, so    */
/* this code should be useful to anyone who needs such code  */
/* on a 32-bit platform.                                     */
/* In particular, note that we don't even #include "ggnfs.h" */
/* so that it can be easily taken out and re-used.           */
/*************************************************************/ 

typedef unsigned long long u64;

typedef void (* MAT_MULT_FUNC_PTR)(u64 *Product, u64 *x, void *P);
int blockLanczos64(u64 *deps, MAT_MULT_FUNC_PTR MultB, MAT_MULT_FUNC_PTR MultB_T, void *P, long n);
/*****************************************************************/ 
/* So here's what to do: Suppose you have a sparse mxn matrix    */
/* 'B' over F_2 with m<n and need some vectors in the kernel.    */
/* Make two functions like these:                                */
/*                                                               */
/*   void multB( long *Product, long *x, void *P)                */
/*   void multBT(long *Product, long *y, void *P)                */
/*                                                               */
/* Which can compute the products:                               */
/* Product <-- Bx,    Product <-- (B^T)y, respectively,          */
/* where 'x' is an nx64 matrix over F_2,                         */
/*       'y' is an mx64 matrix over F_2.                         */
/* The void pointer 'P' is so you can supply a pointer to        */
/* the data defining the matrix as well (i.e., in our case,      */
/* it will point to the relations). It could be NULL if you don't*/
/* need it. Anyway, once you have such a function and need some  */
/* vectors in the kernel, do:                                    */
/*                                                               */
/*  long deps[n];                                                */
/*  myMat_t *P;                                                  */
/*   ...                                                         */
/*  numDeps = blockLanczos(deps, myMultFunction, (void *)P, n);  */
/*                                                               */
/* to get upto 64 dependencies, stored as column vectors in deps.*/
/* Then if numDeps > 0, you will have 'numDeps' vectors in the   */
/* kernel, stored in a bit-packed fashion. For example, the 1st  */
/* one will be:                                                  */
/*  deps[0]&BIT(0), deps[1]&BIT(0),..., deps[n-1]&BIT(0).        */
/*****************************************************************/ 
/* Note: In the case of a matrix from NFS/MPQS, it amounts       */
/* to 'B' being a matrix whose columns are the relations, so that*/
/* we need a dependency among the columns. Although it is        */
/* sometimes tempting to think of the matrix whose rows are the  */
/* relations, it really is much more convenient to do it this    */
/* way.                                                          */
/*****************************************************************/ 
/* It is assumed that 'u64' has 64-bits. It might work if there  */
/* are more, but certainly not if there are less. If there are   */
/* more, you're probably better off adjusting the code anyway.   */
/*****************************************************************/ 

/***** Prototypes for locally used functions. *****/
void multT(u64 *res, u64 *A, u64 *B, long n);
void multS(u64 *D, int *S);
void mult64x64(u64 *res, u64 *A, u64 *B);
void preMult(u64 *A, u64 *B);
void multnx64(u64 *C_n, u64 *A_n, u64 *B, long n);
void getW_S(u64 *Wi, int *Si, u64 *T, int *Si_1);
int  isZeroV(u64 *A, long size);
int  doColumnOps(u64 *A, u64 *B, long n);
int  doColumnOps64(u64 *A, u64 *B, long n);
/**************************************************/


void seedBlockLanczos(long seed)
{ prandseed(seed, 712*seed + 21283, seed^0xF3C91D1A);
}

/**********************************************************************/
int blockLanczos64(u64 *deps, MAT_MULT_FUNC_PTR MultB, MAT_MULT_FUNC_PTR MultB_T, void *P, long n)
/**********************************************************************/
{ u64 *Y=NULL, *X=NULL, *Vi=NULL, *Vi_1=NULL, *Vi_2=NULL, *tmp_n=NULL, *tmp2_n=NULL;
  u64 *V0=NULL, *tmp3_n=NULL, *Z=NULL, *AZ=NULL;
  u64 D[64], E[64], F[64], Wi[64], Wi_1[64], Wi_2[64], T[64], T_1[64], tmp[64];
  u64 U[64], U_1[64], tmp2[64], mask, t64;
  int  Si[64], Si_1[64];
  long i, j, m, isZero, iterations;
  int  errs=0, numDeps=-1, cont, s;

  
  /* Memory allocation: */
  if (!(Y = (u64 *)malloc(n*sizeof(u64))))    errs++;
  if (!(X = (u64 *)malloc(n*sizeof(u64))))    errs++;
  if (!(Vi = (u64 *)malloc(n*sizeof(u64))))   errs++;
  if (!(V0 = (u64 *)malloc(n*sizeof(u64))))   errs++;
  if (!(Vi_1 = (u64 *)malloc(n*sizeof(u64)))) errs++;
  if (!(Vi_2 = (u64 *)malloc(n*sizeof(u64)))) errs++;
  if (!(tmp_n = (u64 *)malloc(n*sizeof(u64)))) errs++;
  if (!(tmp2_n = (u64 *)malloc(n*sizeof(u64)))) errs++;
  if (!(tmp3_n = (u64 *)malloc(n*sizeof(u64)))) errs++;

  if (errs)
    goto SHORT_CIRC_STOP;
  
  /******************************************************************/
  /* Throughout, 'A' means the matrix A := (B^T)B. In fact, all the */
  /* notation is directly from Montgomery's paper, except that I    */
  /* don't bother indexing the D_i, E_i, F_i. They are simply D,E,F.*/
  /******************************************************************/
  
  /* Initialization: */
  for (j=0; j<n; j++) {
    t64 = prand();
    t64 = (t64<<32);
    Y[j] = prand()^t64;  /* Any 32-bit PRNG will do. */
    X[j] = Vi_1[j] = Vi_2[j] = tmp_n[j] = tmp2_n[j] = tmp3_n[j] = 0;
  }
  for (i=0; i<64; i++) {
    Wi[i] = Wi_1[i] = Wi_2[i] = 0;
    T[i] = T_1[i] = U[i] = U_1[i] = tmp[i] = 0;
    Si[i] = Si_1[i] = i; /* Si and Si_1 are both I_64. */
  }
  MultB(tmp_n, Y, P); 
  MultB_T(V0, tmp_n, P);
  memcpy(Vi, V0, n*sizeof(u64));

  /* Prime 'T' for the loop, so that we always have */
  /* T = (Vi^T)A(Vi) and T_1 = (Vi_1^T)A(Vi_1).     */
  
  MultB(tmp2_n, Vi, P);
  MultB_T(tmp_n, tmp2_n, P);  /* tmp_n <-- A*Vi */
  multT(T, Vi, tmp_n, n);      /* T <-- (Vi^T)(tmp_n) = (Vi^T)A(Vi) */
  
  cont = 1;
  i = 0;
  getW_S(Wi, Si, T, Si_1); /* Compute W0 and S0. */
  /* Initialize X <-- (V0)(W0)(V0^T)(V0). */
  multT(tmp, V0, V0, n);    /* tmp <-- (V0^T)(V0). */
  mult64x64(tmp2, Wi, tmp); /* tmp2 <-- (W0)(tmp) = (W0)(V0^T)(V0). */
  multnx64(X, V0, tmp2, n); /* X <-- V0(tmp2). */
  iterations = 0;
  do {
    /* Iteration step. */
    iterations++;

    /********** Compute D_{i+1}. **********/
    MultB(tmp2_n, Vi, P);
    MultB_T(tmp_n, tmp2_n, P); /* tmp_n <-- A*Vi. */
    /* Do not mess with tmp_n after this! We will re-use it below! */

    multT(U, tmp_n, tmp_n, n);  /* U <-- (tmp_n)^T(tmp_n) = (Vi^T)(A^2)(Vi) */
    multS(U, Si);            /* U <-- (Vi^T)(A^2)(Vi)(Si)(Si^T). */
    memcpy(D, U, 64*sizeof(u64)); /* D <-- U. */
    for (j=0; j<64; j++)
      D[j] ^= T[j]; /* D <-- D + (Vi^T)A(Vi). */
    preMult(D, Wi); /* D <-- (Wi)D. */
    for (j=0; j<64; j++)
      D[j] ^= BIT(j); /* D <-- D + I_{64}. */

    /********** Compute E_{i+1}. **********/
    mult64x64(E, Wi_1, T); /* E <-- (Wi_1)(Vi^T)A(Vi). */
    multS(E, Si);          /* E <-- E(Si)(Si^T).       */

    /**************** Compute F_{i+1}. *************/
    /* Note: We should, at this point, always have */
    /* T_1 = (Vi_1^T)A(Vi_1) and                   */
    /* U_1 = (Vi_1^T)A^2(Vi_1)(Si_1)(Si_1^T)       */
    /***********************************************/
    for (j=0; j<64; j++)
      F[j] = U_1[j] ^ T_1[j];
    multS(F, Si);

    mult64x64(tmp, T_1, Wi_1); /* tmp <-- (Vi_1^T)A(Vi_1)(Wi_1). */
    for (j=0; j<64; j++)
      tmp[j] ^= BIT(j); /* tmp <-- tmp + I_64. */
    preMult(tmp, Wi_2);
    preMult(F, tmp);
    /*** Done computing 'F'. */

    /* Finally, do the actual iteration step, putting V_{i+1} in 'tmp_n'. */
    /* Right now, we still have tmp_n = A*Vi, so we should first use it.  */
    
    /* Do tmp_n <-- tmp_n*(Si)(Si^T). */
    mask = 0x0000000000000000ULL;
    for (j=0; j<64; j++) {
      s = Si[j];
      if ((s>=0) && (s<64))
        mask |= BIT(s);
    }
    for (j=0; j<n; j++) 
      tmp_n[j] &= mask;
    
    multnx64(tmp2_n, Vi, D, n);   /* tmp2_n <-- (Vi)D.        */
    for (j=0; j<n; j++)
      tmp_n[j] ^= tmp2_n[j];   /* tmp_n <-- tmp_n + tmp2_n */
    
    multnx64(tmp2_n, Vi_1, E, n); /* tmp2_n <-- (Vi_1)E */
    for (j=0; j<n; j++)
      tmp_n[j] ^= tmp2_n[j];   /* tmp_n <-- tmp_n + tmp2_n */

    multnx64(tmp2_n, Vi_2, F, n); /* tmp2_n <-- (Vi_2)F */
    for (j=0; j<n; j++)
      tmp_n[j] ^= tmp2_n[j];   /* tmp_n <-- tmp_n + tmp2_n */
    /*** Done. We now have tmp_n = V_{i+1}. ***/
   

    i++;
    memcpy(Vi_2, Vi_1, n*sizeof(u64));
    memcpy(Vi_1, Vi, n*sizeof(u64));
    memcpy(Vi, tmp_n, n*sizeof(u64));
    memcpy(Wi_2, Wi_1, 64*sizeof(u64));
    memcpy(Wi_1, Wi, 64*sizeof(u64));
    memcpy(T_1, T, 64*sizeof(u64));
    memcpy(U_1, U, 64*sizeof(u64));
    memcpy(Si_1, Si, 64*sizeof(int));
    
    /******** My `step 3'. ********/
    /* Is (Vi^T)(A)(Vi) == 0 ? */
    MultB(tmp2_n, Vi, P);
    MultB_T(tmp_n, tmp2_n, P);  /* tmp_n <-- A*Vi       */

    multT(T, Vi, tmp_n, n); /* T <-- (Vi^T)A(Vi). */
    getW_S(Wi, Si, T, Si_1);
    
    if (!(isZeroV(T, 64))) {
      /* X <-- X + (Vi)(Wi)(Vi^T)(V_0) */
      multT(tmp, Vi, V0, n); /* tmp <-- (Vi^T)(V0). */
      preMult(tmp, Wi);   /* tmp <-- (Wi)(tmp).  */
      
      multnx64(tmp_n, Vi, tmp, n);   /* tmp_n <-- (Vi)(tmp)    */
      for (j=0; j<n; j++)
        X[j] ^= tmp_n[j]; /* X <-- X + tmp_n. */
    } else {
      cont=0;
      m = i;
    }
    if (iterations % 50 == 0) {
      printTmp("Lanczos: Estimate %1.1lf%% complete...",
                (double)100.0*64.0*iterations/n);  
    }
  } while (cont);
  printf("\nBlock Lanczos used %ld iterations.\n", iterations);

          
  Z = (u64 *)malloc(2*n*sizeof(u64));
  AZ = (u64 *)malloc(2*n*sizeof(u64));
  if (!(Z&&AZ)) {
    fprintf(stderr, "blanczos(): Memory allocation error!\n");
    goto SHORT_CIRC_STOP;
  }
  
  if (isZeroV(Vi, n)) {
    /* Then <X+Y> < ker(A). Later, we will expect AX=0, so do X <-- X+Y. */
    printf("After Block Lanczos iteration, Vm=0.\n");

    for (i=0; i<n; i++)
      X[i] ^= Y[i];
  } else {
    printf("After Block Lanczos iteration, Vm is nonzero. Finishing...\n");
    /* We need more memory (for convenience), so free off what we don't need. */
    free(V0); free(Vi_1); free(Vi_2); free(tmp3_n);
    V0 = Vi_1 = Vi_2 = tmp3_n = NULL;

    /* Construct Z=[ X+Y | Vi] and compute AZ=[ A(X+Y) | AVi ] */
    /* X <-- X+Y, for convenience.                     */
    for (j=0; j<n; j++) {
      X[j] ^= Y[j];
      Z[2*j] = X[j];
      Z[2*j+1] = Vi[j];
    }
    /* Now, compute AZ = [ AX | AVi ]. */
    MultB(tmp2_n, X, P);  
    MultB_T(tmp_n, tmp2_n, P); /* tmp_n <-- AX       */
    for (j=0; j<n; j++) 
      AZ[2*j] = tmp_n[j];
    MultB(tmp2_n, Vi, P);  
    MultB_T(tmp_n, tmp2_n, P); /* tmp_n <-- AVi       */
    for (j=0; j<n; j++) 
      AZ[2*j+1] = tmp_n[j];
    /* Now, AZ should have small rank, and we should be  */
    /* able to do column ops to make zero vectors. Doing */
    /* the same ops to 'Z', we get vectors in ker(A).    */
    doColumnOps(Z, AZ, n);
    /* Now, look for zero columns in AZ. If a column is zero, */
    /* copy the corresponding column of Z into 'X'.           */
    for (i=0; i<n; i++)
      X[i] = 0;
    numDeps=0;
    for (i=0; (i<128) && (numDeps < 64); i++) {
      j=0;
      while ((j<n) && ((AZ[2*j + i/64]&BIT(i%64))==0))
        j++;
      if (j==n) {
        /* Copy column i of Z into column 'numDeps' of X. */
        for (j=0; j<n; j++) {
          if (Z[2*j + i/64]&BIT(i%64))
            X[j] ^= BIT(numDeps);
        }
        numDeps++;
      }
    }
    printf("Found %d dependencies for A=(B^T)B.\n", numDeps);
  }
  j=0;
  while ((j<n)&&(X[j]==0))
    j++;
  if (j==n)
    printf("Probable error: The matrix X is identically zero!!!\n");


  printf("Getting dependencies for original matrix, B...\n");
  /***************************************************/
  /* At this point, we should have AX = (B^T)BX = 0. */
  /* Do simultaneous column ops to BX and X, putting */
  /* BX in RREF, to get vectors in ker(B).           */
  /***************************************************/

  for (i=0; i<n; i++)
    deps[i] = 0;
  numDeps=0;

  MultB(tmp_n, X, P);  /* tmp_n <-- BX. */
  doColumnOps64(X, tmp_n, n);
  for (i=0; i<64; i++) {
    for (j=0, isZero=1; j<n; j++)
      if (tmp_n[j]&BIT(i)) {
        isZero=0;
      }
    if (isZero) {
      for (j=0, isZero=1; j<n; j++) {
        if (X[j]&BIT(i)) {
          deps[j] ^= BIT(numDeps);
          isZero=0;
        }
      }
      if (!(isZero))
        numDeps++;
    }
  }

  if (numDeps) {
    printf("Found %d dependencies for 'B'. Verifying...\n", numDeps);
  } else {
    printf("Some error occurred: all dependencies found seem to be trivial!\n");
    printf("Is Rank(A) = Rank((B^T)B) too small?\n");
    goto SHORT_CIRC_STOP;
  }

  MultB(tmp_n, deps, P);
  for (i=0, isZero=1; (i<n)&&(isZero); i++)
    if (tmp_n[i])
      isZero = 0;
  if (!(isZero))
    printf("Some error occurred: Final product (B)(deps) is nonzero (i=%ld)!\n", i);
  else
    printf("Verified.\n"); 



SHORT_CIRC_STOP:
  
  if (Y != NULL)      free(Y);
  if (X != NULL)      free(X);
  if (Vi != NULL)     free(Vi);
  if (V0 != NULL)     free(V0);
  if (Vi_1 != NULL)   free(Vi_1);
  if (Vi_2 != NULL)   free(Vi_2);
  if (tmp_n != NULL)  free(tmp_n);
  if (tmp2_n != NULL) free(tmp2_n);
  if (tmp3_n != NULL) free(tmp3_n);
  if (Z != NULL)      free(Z);
  if (AZ != NULL)     free(AZ);
  return numDeps;
}

/*******************************************/
void multT(u64 *res, u64 *A, u64 *B, long n)
/*******************************************/
/* Input: A and B are nx64 matrices.       */
/* Output: res <-- (A^T)B, a 64x64 matrix. */
/*******************************************/
{ long i, j;
  u64  bit, thisWord;

  for (i=0; i<64; i++) {
    bit = BIT(i);
    thisWord = 0;
    for (j=0; j<n; j++) {
      if (A[j]&bit)
        thisWord ^= B[j];
    }
    res[i] = thisWord;
  }
}
  


/*********************************************/
void multS(u64 *D, int *S)
/*********************************************/
/* Input: D is a 64x64 matrix,               */
/*        'S' is a subset of the columns,    */
/*        {0,1,2,...,31}.                    */
/* Output: Columns of 'D' not in 'S' are     */
/*         zeroed, and the others unchanged. */
/*********************************************/
{ int i, s;
  u64 mask;

  mask = 0x000000000000000ULL;
  for (i=0; i<64; i++) {
    s = S[i];
    if ((s>=0) && (s<64))
      mask |= BIT(s);
  }
  for (i=0; i<64; i++) 
    D[i] &= mask;
}

/******************************************/
void mult64x64(u64 *res, u64 *A, u64 *B)
/******************************************/
/* Input: 'A' and 'B' are 64x64 matrices. */
/* Output: res <-- A*B.                   */
/******************************************/
{ int i, j;
  u64 thisA, thisRes;

  for (i=0; i<64; i++) {
    thisA = A[i]; thisRes = 0;
    for (j=0; j<64; j++) {
      if (thisA&BIT(j))
        thisRes ^= B[j];
    }
    res[i] = thisRes;
  }
}

/******************************************/
void preMult(u64 *A, u64 *B)
/******************************************/
/* Input: 'A' and 'B' are 64x64 matrices. */
/* Output: A <-- B*A.                     */
/******************************************/
{ u64 res[64];

  mult64x64(res, B, A);
  memcpy(A, res, 64*sizeof(u64));
}
  
/*********************************************/
void multnx64(u64 *C_n, u64 *A_n, u64 *B, long n)
/*********************************************/
/* Input: A_n is an nx64 matrix,             */
/*          B is a 64x64 matrix.             */
/* Output: C_n is the nx64 product (A_n)(B). */
/*********************************************/
/* Remark: There should be a faster way to   */
/* do this, but we'll worry about that later.*/
/*********************************************/
{ long i, j;
  u64  A_row, thisWord, *Aptr;

  Aptr = A_n;
  for (i=0; i<n; i++) {
    A_row = *Aptr++;
    thisWord = 0;
    for (j=0; j<64; j++) {
      if (A_row&BIT(j)) 
        thisWord ^= B[j];
    }
    C_n[i] = thisWord;
  }
}

/*********************************************/ 
int isZeroV(u64 *A, long size)
/*********************************************/ 
{ long i;

  for (i=0; i<size; i++)
    if (A[i]!=0x0000000000000000ULL)
      return 0;
  return 1;
}

/*********************************************/ 
void getW_S(u64 *Wi, int *Si, u64 *T, int *Si_1)
{ u64  M[128], mask, t0, t1;
  int  c[64], i, j, k, sSize, s;
  
  /* Number the columns, with those in Si_1 coming last. */  
  for (sSize=0, mask=0x0000000000000000UL, i=0; i<64; i++) {
    s = Si_1[i];
    if ((s>=0) && (s<64)) {
      mask |= BIT(s);
      sSize++;
    }
  }
  for (i=0, j=0, k=0; i<64; i++) {
    if (mask&BIT(i)) 
      c[64 - sSize + j++] = i;
    else
      c[k++] = i;
  }
  /* S <-- empty set. */
  sSize=0;
  for (i=0; i<64; i++)
    Si[i] = -1;

  /* Set M <-- [ T | I_64] */
  for (i=0; i<64; i++) {
    M[2*i] = T[i];
    M[2*i+1] = BIT(i);
  }
  for (j=0; j<64; j++) {
    for (k=j; (k<64)&& ( (M[2*c[j]]&BIT(c[j]))==0); k++) {
      if (M[2*c[k]]&BIT(c[j])) {
        /* Exchange rows c[j] and c[k] of M */
        t0 = M[2*c[j]];  t1 = M[2*c[j]+1];
        M[2*c[j]] = M[2*c[k]]; M[2*c[j]+1] = M[2*c[k]+1];
        M[2*c[k]] = t0; M[2*c[k]+1] = t1;
      }
    }
    if ((M[2*c[j]]&BIT(c[j]))) {
      Si[sSize++] = c[j];
      /* Now, add row c[j] to other rows as necessary, to zero out */
      /* the rest of column c[j].                                  */
      for (k=0; k<64; k++) {
        if ((k != c[j]) && (M[2*k]&BIT(c[j]))) {
          M[2*k] ^= M[2*c[j]];
          M[2*k+1] ^= M[2*c[j]+1];
        }
      }
    } else {
      for (k=j; (k<64) && ( (M[2*c[j]+1]&BIT(c[j]))==0); k++) {
        if (M[2*c[k]+1]&BIT(c[j])) {
          /* Exchange rows c[j] and c[k] of M */
          t0 = M[2*c[j]];  t1 = M[2*c[j]+1];
          M[2*c[j]] = M[2*c[k]]; M[2*c[j]+1] = M[2*c[k]+1];
          M[2*c[k]] = t0; M[2*c[k]+1] = t1;
        }
      }
      /* Now, add row c[j] to other rows as necessary, to zero out */
      /* the rest of column c[j]+64                                */
      for (k=0; k<64; k++) {
        if ((k != c[j]) && (M[2*k+1]&BIT(c[j]))) {
          M[2*k] ^= M[2*c[j]];
          M[2*k+1] ^= M[2*c[j]+1];
        }
      }
      /* Zero out row c[j]. */
      M[2*c[j]] = 0; M[2*c[j]+1] = 0;
    } /* end 'else'  */
  } /* end 'for j' */
  for (j=0; j<64; j++)
    Wi[j] = M[2*j+1];
}

/*******************************************************/
int doColumnOps(u64 *A, u64 *B, long n)
/*******************************************************/
/* Do column operations (Gaussian elimination) on 'B', */
/* applying the same operations to 'A' as we go.       */
/* Input: 'A' and 'B' are nx128 matrices.               */
/* Output: 'A' and 'B' are changed, as above. i.e., 'B'*/
/*         is in column-reduced echelon form, and the  */
/*         same transformation was applied to 'A'.     */
/*******************************************************/
/* Remark: This is horribly inefficient. There is      */
/* surely a better way to do it. But the easiest is to */
/* just transpose the matrices before and after.       */
/* But it doesn't much matter either way - this guy is */
/* not called from inside any loops, and it's not a    */
/* bottleneck.                                         */
/*******************************************************/
{ u64 *AT, *BT, t;
  long i, j, k, rowSize, col, row;
  int cont;

  rowSize = n/64;
  if (n%64)
    rowSize += 1;
  
  if (!(AT = (u64 *)malloc(128*rowSize*sizeof(u64)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    return -1;
  }
  if (!(BT = (u64 *)malloc(128*rowSize*sizeof(u64)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    free(AT);
    return -1;
  }
  /* Compute AT = A^T and BT = B^T. */
  memset(AT, 0x00, 128*rowSize*sizeof(u64));
  memset(BT, 0x00, 128*rowSize*sizeof(u64));

  for (row=0; row<n; row++) {
    for (col=0; col<64; col++) {
      if (A[2*row]&BIT(col&0x3F))
        AT[col*rowSize + row/64] ^= BIT(row&0x3F);
      if (B[2*row]&BIT(col&0x3F))
        BT[col*rowSize + row/64] ^= BIT(row&0x3F);
    }
    for (col=64; col<128; col++) {
      if (A[2*row + 1]&BIT(col&0x3F))
        AT[col*rowSize + row/64] ^= BIT(row&0x3F);
      if (B[2*row + 1]&BIT(col&0x3F))
        BT[col*rowSize + row/64] ^= BIT(row&0x3F);
    }
  }

  /* Now, apply row ops to get 'BT' into RREF. */
  col=0;
  for (row=0; row<128; row++) {
    /* Is there a row at or after 'row' with a '1' in column 'col'? */
    /* Consider adding a short circuit, testing whole words at a time */
    /* since we are generally expecting small rank for our matrices.  */
    cont=1;
    do {
      k=row;
      while ((k<128)&&((BT[rowSize*k + col/64]&BIT(col&0x3F))==0))
        k++;
      if (k < 128) {
        /* Swap rows 'row' and 'k'. */
        for (i=0; i<rowSize; i++) {
          t = BT[rowSize*row + i]; 
          BT[rowSize*row + i] = BT[rowSize*k + i];
          BT[rowSize*k + i] = t;
          t = AT[rowSize*row + i]; 
          AT[rowSize*row + i] = AT[rowSize*k + i];
          AT[rowSize*k + i] = t;
        }
        /* Use row 'row' to zero out the rest of this column. */
        for (i=0; i<128; i++) {
          if ((BT[i*rowSize + col/64]&BIT(col&0x3F)) && (i != row)) {
            for (j=0; j<rowSize; j++) {
              BT[i*rowSize + j] ^= BT[row*rowSize + j];
              AT[i*rowSize + j] ^= AT[row*rowSize + j];
            }
          }
        } 
        cont=0;
      } else {
        /* Column 'col' is zero at and below row 'row'. */
        cont=1;
      }
      col++;
    } while ((cont) && (col < n));
    if (col >= n)
      break; /* CJM, 10/20/03. */
  }

  /* Compute A = AT^T and B = BT^T. */
  for (i=0; i<2*n; i++)
    A[i] = B[i] = 0;

  for (row=0; row<128; row++) {
    for (col=0; col<n; col++) {
      if (AT[row*rowSize + col/64]&BIT(col&0x3F))
        A[2*col + row/64] ^= BIT(row&0x3F);
      if (BT[row*rowSize + col/64]&BIT(col&0x3F))
        B[2*col + row/64] ^= BIT(row&0x3F);
    }
  }
  free(AT);
  free(BT);
  return 0;
}

/*******************************************************/
int doColumnOps64(u64 *A, u64 *B, long n)
/*******************************************************/
/* Do column operations (Gaussian elimination) on 'B', */
/* applying the same operations to 'A' as we go.       */
/* Input: 'A' and 'B' are nx64 matrices.               */
/* Output: 'A' and 'B' are changed, as above. i.e., 'B'*/
/*         is in column-reduced echelon form, and the  */
/*         same transformation was applied to 'A'.     */
/*******************************************************/
/* Remark: This is horribly inefficient. There is      */
/* surely a better way to do it. But the easiest is to */
/* just transpose the matrices before and after.       */
/* But it doesn't much matter either way - this guy is */
/* not called from inside any loops, and it's not a    */
/* bottleneck.                                         */
/*******************************************************/
{ u64 *AT, *BT, t;
  long i, j, k, rowSize, col, row;
  int cont;

  rowSize = n/64;
  if (n%64)
    rowSize += 1;
  
  if (!(AT = (u64 *)malloc(64*rowSize*sizeof(u64)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    return -1;
  }
  if (!(BT = (u64 *)malloc(64*rowSize*sizeof(u64)))) {
    printf("doColumnOps(): Memory allocation error!\n");
    free(AT);
    return -1;
  }
  /* Compute AT = A^T and BT = B^T. */
  memset(AT, 0x00, 64*rowSize*sizeof(u64));
  memset(BT, 0x00, 64*rowSize*sizeof(u64));

  for (row=0; row<n; row++) {
    for (col=0; col<64; col++) {
      if (A[row]&BIT(col&0x3F))
        AT[col*rowSize + row/64] ^= BIT(row&0x3F);
      if (B[row]&BIT(col&0x3F))
        BT[col*rowSize + row/64] ^= BIT(row&0x3F);
    }
  }

  /* Now, apply row ops to get 'BT' into RREF. */
  col=0;
  for (row=0; row<64; row++) {
    /* Is there a row at or after 'row' with a '1' in column 'col'? */
    /* Consider adding a short circuit, testing whole words at a time */
    /* since we are generally expecting small rank for our matrices.  */
    cont=1;
    do {
      k=row;
      while ((k<64)&&((BT[rowSize*k + col/64]&BIT(col&0x3F))==0))
        k++;
      if (k < 64) {
        /* Swap rows 'row' and 'k'. */
        for (i=0; i<rowSize; i++) {
          t = BT[rowSize*row + i]; 
          BT[rowSize*row + i] = BT[rowSize*k + i];
          BT[rowSize*k + i] = t;
          t = AT[rowSize*row + i]; 
          AT[rowSize*row + i] = AT[rowSize*k + i];
          AT[rowSize*k + i] = t;
        }
        /* Use row 'row' to zero out the rest of this column. */
        for (i=0; i<64; i++) {
          if ((BT[i*rowSize + col/64]&BIT(col&0x3F)) && (i != row)) {
            for (j=0; j<rowSize; j++) {
              BT[i*rowSize + j] ^= BT[row*rowSize + j];
              AT[i*rowSize + j] ^= AT[row*rowSize + j];
            }
          }
        } 
        cont=0;
      } else {
        /* Column 'col' is zero at and below row 'row'. */
        cont=1;
      }
      col++;
    } while ((cont) && (col < n));
    if (col >= n)
      break; /* CJM, 10/20/03. */
  }
  /* Compute A = AT^T and B = BT^T. */
  for (i=0; i<n; i++)
    A[i] = B[i] = 0;

  for (row=0; row<64; row++) {
    for (col=0; col<n; col++) {
      if (AT[row*rowSize + col/64]&BIT(col&0x3F))
        A[col] ^= BIT(row);
      if (BT[row*rowSize + col/64]&BIT(col&0x3F))
        B[col] ^= BIT(row);
    }
  }
  free(AT);
  free(BT);
  return 0;
}




