/* mpqs.c
  Written by T. Kleinjung. 
  6/13/04: Hacked up for inclusion in GGNFS by Chris Monico.

  Copyright (C) 2001 Jens Franke, T. Kleinjung.
  This file is part of gnfs4linux, distributed under the terms of the 
  GNU General Public Licence and WITHOUT ANY WARRANTY.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <math.h>
#include <gmp.h>
#include "lasieve-asm.h"
#include "../lasieve.h"

#define ull unsigned long long
#define uchar unsigned char
#ifdef MPQS_STAT
extern ull stat_asm_eval, stat_asm_td;
extern ulong stat_asm_div, stat_final_mulmod;
#endif

#define L1BITS               14

#define MPQS_MAX_FBSIZE      512
#define MPQS_MIN_EXCESS      10
#define MPQS_MAX_ADIV_ALL    6
#define MPQS_SIEVELEN        (1<<L1BITS)
#define MPQS_REL_ENTRIES     32
#define MPQS_TD_MAX_NDIV     (MPQS_REL_ENTRIES-5)
#define MPQS_MAX_NRELS       256        /* ??? */
#define MPQS_MAX_NPRELS      2048       /* ??? */
#define MPQS_MAX_NCRELS      256        /* ??? */
#define MPQS_MAX_NRELS_BUF   256        /* ??? */
#define MPQS_MIN_RELBUFFER   16
#define MPQS_GAUSS_MAX       512
#define MPQS_MAX_FACTORS     5
#define MPQS_MAX_NPRIMES     1024
#define MPQS_HASH_OVERFLOW   65535
#define MPQS_MULT_NTESTPR    25 /* prime(MPQS_MULT_NTESTPR)<2^8 */
#define MPQS_MULT_NCAND      8  /* largest candidate<2^8 */
#define MPQS_FB_MAXPRIME     4096       /* used in sieve and final */

#ifndef _MSC_VER
#define NAME(_a) asm(_a)
#else
#define NAME(_a)
#endif

/* common with asm functions */
ushort mpqs_nFBk_1 NAME("mpqs_nFBk_1");
ushort mpqs_td_begin NAME("mpqs_td_begin");
ushort mpqs_sievebegin NAME("mpqs_sievebegin");
ushort mpqs_FB_inv_info[4 * MPQS_MAX_NPRIMES] NAME("mpqs_FB_inv_info");
ushort mpqs_nFBk NAME("mpqs_nFBk");
ushort mpqs_FBk[3] NAME("mpqs_FBk");  /* 3*5*7*11 too big */
uchar mpqs_FB_log[MPQS_MAX_FBSIZE] NAME("mpqs_FB_log");
ushort mpqs_nFB NAME("mpqs_nFB");
ushort mpqs_nAdiv_total NAME("mpqs_nAdiv_total");
ushort mpqs_Adiv_all[MPQS_MAX_ADIV_ALL] NAME("mpqs_Adiv_all");
ulong mpqs_FBk_inv[3] NAME("mpqs_FBk_inv");
ulong mpqs_FB_A_inv[MPQS_MAX_ADIV_ALL] NAME("mpqs_FB_A_inv");

/* end */

static int mpqs_isinit = 0;
static mpz_t mpqs_gmp_help, mpqs_dummy, mpqs_help;
static mpz_t mpqs_N, mpqs_kN, mpqs_factors[MPQS_MAX_FACTORS];
static uchar mpqs_f_status[MPQS_MAX_FACTORS];
static ulong mpqs_multiplier, mpqs_nfactors;
static double mpqs_complexity;
static ushort mpqs_prime_table[MPQS_MAX_NPRIMES];
static ushort mpqs_prime_sqrthelp[MPQS_MAX_NPRIMES];
static ushort mpqs_nAdiv, mpqs_accept;
static uchar mpqs_Adiv_log[MPQS_MAX_ADIV_ALL];;
static ushort mpqs_pmax;
static ushort mpqs_Adiv[MPQS_MAX_ADIV_ALL];
static ushort mpqs_Adiv_sqrt[MPQS_MAX_ADIV_ALL], mpqs_Adiv_all_sqrt[MPQS_MAX_ADIV_ALL];
static ushort mpqs_Adiv_start1[MPQS_MAX_ADIV_ALL], mpqs_Adiv_start2[MPQS_MAX_ADIV_ALL];
static uchar mpqs_Adiv_active[MPQS_MAX_ADIV_ALL];
static uchar mpqs_A_mask[(1 << MPQS_MAX_ADIV_ALL)];
static ushort mpqs_nA;
static short mpqs_A_index, mpqs_B_index;
static ushort mpqs_SI_add[MPQS_MAX_ADIV_ALL][MPQS_MAX_FBSIZE];
static ushort mpqs_2A_all_inv[MPQS_MAX_FBSIZE];
static ushort mpqs_Adiv_SI_add[MPQS_MAX_ADIV_ALL][MPQS_MAX_ADIV_ALL];
static long long mpqs_A, mpqs_B, mpqs_C, mpqs_2B, mpqs_Bi[MPQS_MAX_ADIV_ALL];
static ulong mpqs_disp;
static ushort mpqs_nsurvivors;
static ulong mpqs_nrels, mpqs_nlp, mpqs_nsp, mpqs_excess;
static uchar mpqs_logbound;
static ushort mpqs_nprels, mpqs_ncrels;
static ushort **mpqs_relations, **mpqs_part_rels, **mpqs_comb_rels;
static ushort **mpqs_rel_buffer;
static ushort mpqs_rel_hash_table[256][16];
static ushort mpqs_hash_table[128][16];
static ulong mpqs_hash_index[128][16];
static ulong mpqs_lp[128 * 15];
static uchar *mpqs_rel_active;  /* auf bitarray umstellen !!! */
static ushort mpqs_nrels_active, mpqs_nlp_active;
static ushort mpqs_lp_count[128 * 16 + MPQS_MAX_FBSIZE + 3 + MPQS_MAX_ADIV_ALL];
static ulong **mpqs_gauss_row;
static uchar mpqs_sol[MPQS_GAUSS_MAX];
static short mpqs_gauss_c[MPQS_GAUSS_MAX], mpqs_gauss_d[MPQS_GAUSS_MAX];
static long mpqs_gauss_m, mpqs_gauss_n, mpqs_gauss_n32, mpqs_gauss_k;
static short mpqs_exp[128 * 16 + MPQS_MAX_FBSIZE + 3 + MPQS_MAX_ADIV_ALL];
static mpz_t mpqs_sq1, mpqs_sq2;
static ushort mpqs_zero1, mpqs_zero2;
static double mpqs_kN_dbl;
static ull mpqs_kN_64, mpqs_A_inv_64;

ushort mpqs_FB[2 * MPQS_MAX_FBSIZE];
ushort mpqs_FB_start[2 * MPQS_MAX_FBSIZE];
ulong mpqs_sievelen;
ulong mpqs_FB_inv[MPQS_MAX_FBSIZE];
uchar *mpqs_sievearray;

uchar mpqs_256_inv_table[128] = {
  1, 171, 205, 183, 57, 163, 197, 239,
  241, 27, 61, 167, 41, 19, 53, 223,
  225, 139, 173, 151, 25, 131, 165, 207,
  209, 251, 29, 135, 9, 243, 21, 191,
  193, 107, 141, 119, 249, 99, 133, 175,
  177, 219, 253, 103, 233, 211, 245, 159,
  161, 75, 109, 87, 217, 67, 101, 143,
  145, 187, 221, 71, 201, 179, 213, 127,
  129, 43, 77, 55, 185, 35, 69, 111,
  113, 155, 189, 39, 169, 147, 181, 95,
  97, 11, 45, 23, 153, 3, 37, 79,
  81, 123, 157, 7, 137, 115, 149, 63,
  65, 235, 13, 247, 121, 227, 5, 47,
  49, 91, 125, 231, 105, 83, 117, 31,
  33, 203, 237, 215, 89, 195, 229, 15,
  17, 59, 93, 199, 73, 51, 85, 255
};

#ifdef MPQS_STAT
ulong stat_mpqs_nsieves, stat_mpqs_nsurvivors, stat_mpqs_ntrials, stat_mpqs_ndiv;
#endif

/*******************************************************/
static ull mpqs_inv_64(ull a)
/*******************************************************/
{ ull inv, h;

  inv = (ull) mpqs_256_inv_table[(a & 0xff) >> 1];
  h = a * inv;
  h &= 0xff00;
  h *= inv;
  inv -= h;
  h = a * inv;
  h &= 0xffff0000;
  h *= inv;
  inv -= h;
  h = a * inv;
  h &= 0xffffffff00000000ULL;   /* CJM: ULL added for gcc; other compilers may differ. */
  h *= inv;
  inv -= h;
  /*if (inv*a!=1ULL) Schlendrian("mpqs_inv_64"); */
  return inv;
}

static uchar mpqs_jacobi_tab2[8] = { 0, 0, 0, 1, 0, 1, 0, 0 };

/*******************************************************/
static long mpqs_jacobi(ushort a, ushort b)
/*******************************************************/
/* always gcd(a,b)=1 and 0<a<b */
{ long e, l, m, n, r;

  m = (long) a;
  n = (long) b;
  if (!(n & 1))
    Schlendrian("mpqs_jacobi: b even\n");
  e = 0;
  if (!(m & 1)) {
    l = mpqs_jacobi_tab2[n & 7];
    do {
      m /= 2;
      e += l;
    } while (!(m & 1));
  }
  while (m > 1) {
    r = n - m;
    if (r > 0) {
      l = m & n;
      l >>= 1;
      e += l;
      n = m;
      m = r;
    } else {
      m = -r;
    }
    if (m & 1)
      continue;
    l = mpqs_jacobi_tab2[n & 7];
    do {
      m /= 2;
      e += l;
    } while (!(m & 1));
  }
  if (!m)
    Schlendrian("jacobi\n");
  e = e & 1;
  e = 1 - 2 * e;
  return e;
}

/*******************************************************/
static long mpqs_jacobi0(ushort a, ushort b)
/*******************************************************/
/* always gcd(a,b)=1 and 0<a<b */
{ long e, l, m, n;

  m = (long) a;
  n = (long) b;
  if (!(n & 1))
    Schlendrian("mpqs_jacobi: b even\n");
  e = 0;
  if (!(m & 1)) {
    l = mpqs_jacobi_tab2[n & 7];
    do {
      m /= 2;
      e += l;
    } while (!(m & 1));
  }
  while (m > 1) {
    l = m;
    m = n;
    n = l;
    l = m & n;
    l >>= 1;
    e += l;
    l = m / n;
    m -= l * n;
    if (m & 1)
      continue;
    l = mpqs_jacobi_tab2[n & 7];
    do {
      m /= 2;
      e += l;
    } while (!(m & 1));
  }
  if (!m)
    Schlendrian("jacobi\n");
  e = e & 1;
  e = 1 - 2 * e;
  return e;
}
  
/*******************************************************/
static ushort mpqs_invert(ushort a, ushort p)
/*******************************************************/
{ ulong v1 = 0, v2 = 1;
  ulong b = (ulong) a, pp = (ulong) p, q;

  while (b > 1) {
    pp -= b;
    v1 += v2;
    if (pp >= b) {
      pp -= b;
      v1 += v2;
      if (pp >= b) {
        q = pp / b;
        pp %= b;
        v1 += q * v2;
      }
    }
    if (pp <= 1) {
      v2 = p - v1;
      break;
    }
    b -= pp;
    v2 += v1;
    if (b >= pp) {
      b -= pp;
      v2 += v1;
      if (b >= pp) {
        q = b / pp;
        b %= pp;
        v2 += q * v1;
      }
    }
  }
/*  if ((v2*a-1)%(ulong)p) complain("invert");*/
  return v2;
}

/*******************************************************/
static ushort mpqs_powmod(ushort a, ushort e, ushort p)
/*******************************************************/
{ ushort ex = e;
  ulong aa, res;

  if (!ex)
    return 1;
  aa = (ulong) a;
  res = 1;
  while (1) {
    if (ex & 1) {
      res *= aa;
      res %= (ulong) p;
    }
    ex >>= 1;
    if (!ex)
      return (ushort) res;
    aa *= aa;
    aa %= (ulong) p;
  }
  return 0;                     /* never reached */
}

/*******************************************************/
static ushort mpqs_sqrt_init(ushort p)
/*******************************************************/
{ ushort e, u, i, j;
  ulong g, b;

  if (p & 2)
    return 0;
  if (p & 4)
    return mpqs_powmod(2, (p - 1) >> 2, p);
  e = 0;
  u = p - 1;
  while (!(u & 1)) {
    e++;
    u >>= 1;
  }
  for (i = 2; i < p; i++) {
    g = mpqs_powmod(i, u, p);
    if (g == 1)
      continue;
    b = g;
    for (j = 0; j < e - 1; j++) {
      b *= b;
      b %= (ulong) p;
    }
    if (b == p - 1)
      break;
  }
  if (i >= p)
    Schlendrian("mpqs_sqrt_init\n");
  return g;
}

/*******************************************************/
static ushort mpqs_sqrt(ushort a, ushort p, ushort help)
/*******************************************************/
/* a!=0, a<p, p prime, p!=2, (a/p)=1                   */
/*******************************************************/
{ ushort e, u, i, l;
  ulong b, g, r, k;

  if (p & 2)
    return mpqs_powmod(a, (p + 1) >> 2, p);
  if (p & 4) {
    b = mpqs_powmod(a, (p + 3) >> 3, p);
    g = b * b;
    g %= (ulong) p;
    if (g == a)
      return b;
    b *= (ulong) help;
    b %= (ulong) p;
    return b;
  }
  e = 0;
  u = p - 1;
  while (!(u & 1)) {
    e++;
    u >>= 1;
  }
  g = help;
#if 1
  r = mpqs_powmod(a, u >> 1, p);
  k = r * r;
  k %= (ulong) p;
  k *= a;
  k %= (ulong) p;
  r *= a;
  r %= (ulong) p;
#else
  r = mpqs_powmod(a, (u + 1) / 2, p);
  k = mpqs_powmod(a, u, p);
#endif
  while (k != 1) {
    b = k;
    for (i = 0; i < e && b != 1; i++) {
      b *= b;
      b %= (ulong) p;
    }
    for (l = i + 1; l < e; l++) {
      g *= g;
      g %= (ulong) p;
    }
    r *= g;
    r %= (ulong) p;
    g *= g;
    g %= (ulong) p;
    k *= g;
    k %= (ulong) p;
    e = i;
  }
  return (ushort) r;
}

/*******************************************************/
static ulong mpqs_inv(ushort a)
/*******************************************************/
{ ulong inv, h;

  inv = (ulong) mpqs_256_inv_table[(a & 0xff) >> 1];
  h = (ulong) a *inv;

  h &= 0xff00;
  h *= inv;
  inv -= h;
  h = (ulong) a *inv;

  h &= 0xffff0000;
  h *= inv;
  inv -= h;
  if (inv * (ulong) a != 1ULL)
    Schlendrian("mpqs_inv");
  return inv;
}


/*******************************************************/
void mpqs_init()
/*******************************************************/
{ long i, j;
  ulong p, add, d;

  mpz_init(mpqs_N);
  mpz_init(mpqs_kN);
  mpz_init(mpqs_dummy);
  mpz_init(mpqs_gmp_help);
  mpz_init(mpqs_help);
  mpz_init(mpqs_sq1);
  mpz_init(mpqs_sq2);
  for (i = 0; i < MPQS_MAX_FACTORS; i++)
    mpz_init(mpqs_factors[i]);
  mpqs_sievearray = (uchar *) xmalloc(MPQS_SIEVELEN * sizeof(uchar));
  mpqs_relations = (ushort **) xmalloc(MPQS_MAX_NRELS * sizeof(ushort *));
  mpqs_relations[0] =
    (ushort *) xmalloc(MPQS_MAX_NRELS * MPQS_REL_ENTRIES * sizeof(ushort));
  for (i = 1; i < MPQS_MAX_NRELS; i++)
    mpqs_relations[i] = mpqs_relations[i - 1] + MPQS_REL_ENTRIES;
  mpqs_part_rels = (ushort **) xmalloc(MPQS_MAX_NPRELS * sizeof(ushort *));
  mpqs_part_rels[0] =
    (ushort *) xmalloc(MPQS_MAX_NPRELS * MPQS_REL_ENTRIES * sizeof(ushort));
  for (i = 1; i < MPQS_MAX_NPRELS; i++)
    mpqs_part_rels[i] = mpqs_part_rels[i - 1] + MPQS_REL_ENTRIES;
  mpqs_comb_rels = (ushort **) xmalloc(MPQS_MAX_NCRELS * sizeof(ushort *));
  mpqs_comb_rels[0] =
    (ushort *) xmalloc(MPQS_MAX_NCRELS * 2 * MPQS_REL_ENTRIES *
                       sizeof(ushort));
  for (i = 1; i < MPQS_MAX_NCRELS; i++)
    mpqs_comb_rels[i] = mpqs_comb_rels[i - 1] + 2 * MPQS_REL_ENTRIES;
  mpqs_rel_buffer =
    (ushort **) xmalloc(MPQS_MAX_NRELS_BUF * sizeof(ushort *));
  mpqs_rel_buffer[0] =
    (ushort *) xmalloc(MPQS_MAX_NRELS_BUF * MPQS_REL_ENTRIES *
                       sizeof(ushort));
  for (i = 1; i < MPQS_MAX_NRELS_BUF; i++)
    mpqs_rel_buffer[i] = mpqs_rel_buffer[i - 1] + MPQS_REL_ENTRIES;

  mpqs_rel_active = (uchar *) xmalloc(MPQS_MAX_NRELS * sizeof(uchar));
  mpqs_gauss_row = (ulong **) xmalloc(MPQS_GAUSS_MAX * sizeof(ulong *));
  mpqs_gauss_row[0] =
    (ulong *) xmalloc(MPQS_GAUSS_MAX * MPQS_GAUSS_MAX / 32 * sizeof(ulong));
#if 0
  prime_table_init();
  for (i = 0; i < MPQS_MAX_NPRIMES; i++) {
    mpqs_prime_table[i] = (ushort) get_next_prime();
    mpqs_prime_sqrthelp[i] = mpqs_sqrt_init(mpqs_prime_table[i]);
  }
#else
  mpqs_prime_table[0] = 2;
  mpqs_prime_table[1] = 3;
  mpqs_prime_table[2] = 5;
  mpqs_prime_sqrthelp[2] = mpqs_sqrt_init(5);
  i = 3;
  p = 7;
  add = 4;
  while (i < MPQS_MAX_NPRIMES) {
    for (j = 2; j < i; j++) {
      d = (ulong) mpqs_prime_table[j];
      if (d * d >= p)
        break;
      if (p % d == 0)
        break;
    }
    if (d * d > p) {
      mpqs_prime_table[i] = (ushort) p;
      mpqs_prime_sqrthelp[i] = mpqs_sqrt_init(p);
      i++;
    }
    p += add;
    add = 6 - add;
  }
#endif

  mpqs_isinit = 1;
}

static ushort mpqs_multiplier_cand[4][MPQS_MULT_NCAND] = {
  {1, 17, 33, 41, 57, 65, 73, 89},
  {3, 11, 19, 35, 43, 51, 59, 67},
  {5, 13, 21, 29, 37, 53, 61, 69},
  {7, 15, 23, 31, 39, 47, 55, 71},
};

/*******************************************************/
static void mpqs_choose_multiplier()
/*******************************************************/
{ ushort Nmod8, mult, p, mm;
  ushort residue[MPQS_MULT_NTESTPR];
  double value[MPQS_MULT_NCAND], v, vp, vmin, dn;
  long i, j;

  for (i = 0; i < MPQS_MULT_NTESTPR; i++)
    residue[i] =
      (ushort) mpz_mod_ui(mpqs_dummy, mpqs_N, (ulong) mpqs_prime_table[i]);
  Nmod8 = (ushort) mpz_mod_ui(mpqs_dummy, mpqs_N, 8);
  dn = log(mpz_get_d(mpqs_N)) / log(2.);
  dn = 100.9 - dn;
  if (dn > 7)
    mm = 128;
  else
    mm = (ushort) (exp(dn * log(2)));
  for (j = 0; j < MPQS_MULT_NCAND; j++) {
    mult = mpqs_multiplier_cand[Nmod8 / 2][j];
    if (mult <= mm) {
      v = log((double) mult) - 1.5 * log(2);
      for (i = 1; i < MPQS_MULT_NTESTPR; i++) {
        p = mpqs_prime_table[i];
        vp = log((double) p) / ((double) p);
        if (mult % p) {
          if (mpqs_jacobi((mult * residue[i]) % p, p) == 1)
            v -= vp;
          else
            v += vp;
        }
      }
    } else
      v = 1000;
    value[j] = v;
  }
  vmin = value[0];
  for (j = 1; j < MPQS_MULT_NCAND; j++)
    if (value[j] < vmin)
      vmin = value[j];
  for (j = 0; j < MPQS_MULT_NCAND; j++)
    if (value[j] == vmin)
      break;
  if (j >= MPQS_MULT_NCAND)
    complain("mpqs_choose_multiplier.vmin\n");
  mpqs_complexity = (vmin + log(mpz_get_d(mpqs_N))) / log(2.);
  mpqs_multiplier = mpqs_multiplier_cand[Nmod8 / 2][j];
#ifdef MPQS_STAT
  printf("%ld ", mpqs_multiplier);
#endif
  mpz_mul_ui(mpqs_kN, mpqs_N, mpqs_multiplier);
  mpz_fdiv_q_2exp(mpqs_dummy, mpqs_kN, 32);
  mpqs_kN_64 = (ull) mpz_get_ui(mpqs_dummy);
  mpqs_kN_64 <<= 32;
  mpqs_kN_64 += (ull) mpz_get_ui(mpqs_kN);
}

/*******************************************************/
static void mpqs_choose_multiplier0()
/*******************************************************/
{ ulong Nmod8;

  Nmod8 = mpz_mod_ui(mpqs_dummy, mpqs_N, 8);
  mpqs_multiplier = Nmod8;      /* vorlufig !!!!!!!!!!!!!!!!!!!!!! */
  mpz_mul_ui(mpqs_kN, mpqs_N, mpqs_multiplier);
  mpz_fdiv_q_2exp(mpqs_dummy, mpqs_kN, 32);
  mpqs_kN_64 = (ull) mpz_get_ui(mpqs_dummy);
  mpqs_kN_64 <<= 32;
  mpqs_kN_64 += (ull) mpz_get_ui(mpqs_kN);
}

#if 1
static ushort mpqs_param[12][6] = {
  {40, 3, 2, 4, 11, 16},
  {50, 3, 2, 4, 12, 16},
  {60, 3, 3, 4, 15, 16},
  {70, 3, 3, 5, 14, 16},
  {80, 3, 3, 5, 14, 16},
  {90, 3, 3, 5, 15, 20},
  {110, 3, 3, 5, 17, 20},
  {120, 3, 3, 5, 19, 20},
  {140, 3, 4, 6, 18, 30},
  {140, 3, 4, 6, 20, 40},
  {160, 3, 4, 6, 21, 50},
  {180, 4, 4, 6, 23, 70}
};

/* vor retry-implementierung
static ushort mpqs_param[12][6]={
 { 40, 3, 2, 4, 11, 16},
 { 50, 3, 2, 4, 12, 16},
 { 60, 3, 3, 4, 15, 16},
 { 70, 3, 3, 5, 14, 16},
 { 80, 3, 3, 5, 14, 16},
 { 90, 3, 3, 5, 15, 20},
 { 110, 3, 3, 5, 17, 20},
 { 120, 3, 3, 5, 19, 20},
 { 140, 3, 4, 6, 18, 30},
 { 140, 3, 4, 6, 20, 40},
 { 160, 3, 4, 6, 21, 50},
 { 180, 4, 4, 6, 23, 70}
};
*/
#else
static ushort mpqs_param[10][6] = {
  {60, 3, 3, 4, 13, 16},
  {70, 3, 3, 5, 14, 16},
  {80, 3, 3, 5, 14, 16},
  {90, 3, 3, 5, 15, 20},
  {110, 3, 3, 5, 16, 20},
  {120, 3, 3, 5, 18, 20},
  {130, 3, 4, 6, 18, 30},
  {130, 3, 4, 6, 19, 30},
  {150, 3, 4, 6, 20, 50},
  {180, 4, 4, 6, 23, 70}
};
#endif

/*******************************************************/
static void mpqs_choose_parameter(long retry)
/*******************************************************/
{ long n, r;

#if 1
  n = (long) (mpqs_complexity + 0.5);
  if (n > 96)
    n = 96;
  if (n < 51)
    n = 51;
  n = (n + 3) / 4;
  n -= 13;                      /* 0<=n<=11 */
#else
  n = (long) (mpqs_complexity + 0.5);
  if (n > 96)
    n = 96;
  if (n < 59)
    n = 59;
  n = (n + 3) / 4;
  n -= 15;                      /* 0<=n<=9 */
#endif

  r = retry;
  if ((r) && (n < 9)) {
    n++;
    r--;
  }
  mpqs_nFB = mpqs_param[n][0];
  if (r)
    mpqs_nFB += r * mpqs_nFB / 4;
  if (mpqs_nFB >= MPQS_MAX_FBSIZE)
    mpqs_nFB = MPQS_MAX_FBSIZE - 1;
  mpqs_sievebegin = mpqs_param[n][1];
  if (r)
    mpqs_sievebegin += r;
  mpqs_nAdiv = mpqs_param[n][2];
  mpqs_nAdiv_total = mpqs_param[n][3];
  mpqs_accept = mpqs_param[n][4];
  mpqs_td_begin = mpqs_param[n][5];

  mpqs_sievelen = MPQS_SIEVELEN;        /* !!! */
  mpqs_disp = mpqs_sievelen / 2;
  mpqs_nfactors = 0;
  mpqs_nrels = 0;
  mpqs_nlp = 0;
  mpqs_nprels = 0;
  mpqs_ncrels = 0;
  for (n = 0; n < 128; n++)
    mpqs_hash_table[n][0] = 0;
  for (n = 0; n < 256; n++)
    mpqs_rel_hash_table[n][0] = 0;
}

/*******************************************************/
static void mpqs_generate_FB()
/*******************************************************/
{ ushort *fb, p, rest, i, nfb;

  fb = mpqs_FB;
  nfb = 0;
  i = 0;
  mpqs_nFBk = 0;
  *fb++ = 2;
  *fb++ = 1;
  nfb++;
  for (i = 1; i < MPQS_MAX_NPRIMES; i++) {
    p = mpqs_prime_table[i];
    if (p > MPQS_FB_MAXPRIME)
      break;
    rest = (ushort) mpz_mod_ui(mpqs_dummy, mpqs_kN, p); /* ca 5M */
    if (rest) {
      if (mpqs_jacobi(rest, p) == 1) {  /* ca 10M */
        *fb++ = p;
        *fb++ = mpqs_sqrt(rest, p, mpqs_prime_sqrthelp[i]);     /* ca 13M */
        nfb++;
        if (nfb >= mpqs_nFB) {
          break;
        }
      }
    } else {
      if (mpqs_multiplier % (ulong) p)
        complain("mpqs: N has small divisor: %u\n", p);
      mpqs_FBk[mpqs_nFBk++] = p;
    }
  }
  mpqs_nFB = nfb;
  mpqs_nFBk_1 = mpqs_nFBk + 1;
  mpqs_nsp = 1 + mpqs_nFB + mpqs_nFBk;
  mpqs_pmax = mpqs_FB[2 * (nfb - 1)];
  mpqs_FB[2 * nfb] = mpqs_sievelen;
}

/*******************************************************/
static int mpqs_SI_init()
/*******************************************************/
{ double d;
  long a, i, j, k, l, n;
  long long prod;
  ushort inv, p, *fb = mpqs_FB;
  double A_value[20], A_div_log[MPQS_MAX_ADIV_ALL];     /* maximal number of A's */
  double v, vmax;

  d = mpz_get_d(mpqs_kN);
  mpqs_kN_dbl = d;
  d *= 8.;
  d = sqrt(d);
  d /= (double) mpqs_sievelen;
  d = log(d);
  d /= (double) mpqs_nAdiv;
  d = exp(d);
  a = (long) d;
/*  if (a>=mpqs_pmax) { printf("%ld %ld ",a,mpqs_pmax); return -1; }*/
  if (a >= mpqs_pmax)
    a = mpqs_pmax - 1;
  for (i = 0; i < mpqs_nFB; i++)
    if (mpqs_FB[2 * i] > a)
      break;
  i -= mpqs_nAdiv_total / 2;
  if (i < 1)
    i = 1;                      /* first prime >2 */
  if (mpqs_FB[2 * i] < 3)
    return -2;
  if (i + mpqs_nAdiv_total > mpqs_nFB)  /*return -3; */
    i = mpqs_nFB - mpqs_nAdiv_total - 1;
  for (j = 0; j < mpqs_nAdiv_total; j++) {
    mpqs_Adiv_all[j] = mpqs_FB[2 * (i + j)];
    mpqs_Adiv_all_sqrt[j] = mpqs_FB[2 * (i + j) + 1];
  }
  for (j = i + mpqs_nAdiv_total; j < mpqs_nFB; j++) {
    mpqs_FB[2 * (j - mpqs_nAdiv_total)] = mpqs_FB[2 * j];
    mpqs_FB[2 * (j - mpqs_nAdiv_total) + 1] = mpqs_FB[2 * j + 1];
  }
  mpqs_nFB -= mpqs_nAdiv_total;
  if (i < mpqs_sievebegin)
    mpqs_sievebegin = i;
#if 1
  p = mpqs_FB[2];
  mpqs_FB_inv[1] = mpqs_inv(p);
  mpqs_FB_inv_info[2] = p;
  mpqs_FB_inv_info[3] = p;
  mpqs_FB_inv_info[6] = (ushort) mpqs_FB_inv[1];
  mpqs_FB_inv_info[7] = (ushort) mpqs_FB_inv[1];
  if (mpqs_td_begin & 1)
    complain("mpqs_td_begin is odd\n");
  for (j = 2; j < mpqs_td_begin; j += 2) {
    p = mpqs_FB[2 * j];
    mpqs_FB_inv[j] = mpqs_inv(p);
    mpqs_FB_inv_info[4 * j] = p;
    mpqs_FB_inv_info[4 * j + 1] = p;
    mpqs_FB_inv_info[4 * j + 4] = (ushort) mpqs_FB_inv[j];
    mpqs_FB_inv_info[4 * j + 5] = (ushort) mpqs_FB_inv[j];
    p = mpqs_FB[2 * (j + 1)];
    mpqs_FB_inv[j + 1] = mpqs_inv(p);
    mpqs_FB_inv_info[4 * j + 2] = p;
    mpqs_FB_inv_info[4 * j + 3] = p;
    mpqs_FB_inv_info[4 * j + 6] = (ushort) mpqs_FB_inv[j + 1];
    mpqs_FB_inv_info[4 * j + 7] = (ushort) mpqs_FB_inv[j + 1];
  }
  for (; j < mpqs_nFB; j++)
    mpqs_FB_inv[j] = mpqs_inv(mpqs_FB[2 * j]);
  for (j = 0; j < mpqs_nFBk; j++)
    mpqs_FBk_inv[j] = mpqs_inv(mpqs_FBk[j]);
  for (j = 0; j < mpqs_nAdiv_total; j++)
    mpqs_FB_A_inv[j] = mpqs_inv(mpqs_Adiv_all[j]);
#endif
/* compute log-approximations */
  d = mpz_get_d(mpqs_kN);
  d /= 8.;
  d = sqrt(d);
  d *= (double) mpqs_sievelen;
  d = log(d);
  d -= log((double) (1 << mpqs_accept));
  mpqs_logbound = 128;          /* ??? */
  d = (double) (mpqs_logbound) / d;
  for (i = 0; i < mpqs_nFB; i++) {
    p = mpqs_FB[2 * i];
    mpqs_FB_log[i] = (uchar) (0.5 + d * log((double) p));
  }
  for (i = 0; i < mpqs_nAdiv_total; i++) {
    p = mpqs_Adiv_all[i];
    mpqs_Adiv_log[i] = (uchar) (0.5 + d * log((double) p));
  }

/* compute mask for choice of A-divisors */
  n = 1 << mpqs_nAdiv_total;
  for (i = 0; i < n; i++)
    mpqs_A_mask[i] = 0;
  for (i = 0; i < mpqs_nAdiv_total; i++) {
    k = 1 << i;
    for (j = k; j < n; j += (2 * k))
      for (l = 0; l < k; l++)
        mpqs_A_mask[j + l]++;
  }                             /* now mpqs_A_mask[i] contains the number of ones of i (binary) */
  j = 0;
  for (i = 0; i < n; i++)
    if (mpqs_A_mask[i] == mpqs_nAdiv) {
      mpqs_A_mask[j] = i;
      j++;
    }
/* ensure that |C|<2^64 */
  for (i = 0; i < mpqs_nAdiv_total; i++)
    A_div_log[i] = log((double) (mpqs_Adiv_all[i])) / log(2.);
  d = log(mpqs_kN_dbl) / log(2.);
  for (i = 0; i < j; i++) {
    v = 0;
    for (k = 0; k < mpqs_nAdiv_total; k++)
      if (mpqs_A_mask[i] & (1 << k))
        v += A_div_log[k];
    if (d - v > 63.95) {
      j--;                      /* printf(":"); */
      mpqs_A_mask[i] = mpqs_A_mask[j];
      i--;
    }
  }

  mpqs_nA = j;
  mpqs_A_index = -1;
  mpqs_B_index = (1 << mpqs_nAdiv) - 1;

  prod = 1;
  d = 1.;
  for (i = 0; i < mpqs_nAdiv_total; i++) {
    prod *= (long long) (mpqs_Adiv_all[i]);
    d *= (double) (mpqs_Adiv_all[i]);
  }
  if (d > 9223372036854775807.)
    return -4;
  for (i = 1; i < mpqs_nFB; i++) {
    p = fb[2 * i];
    inv = (ushort) (prod % (long long) p);
    inv = mpqs_invert(inv, p);
    inv <<= 1;
    if (inv >= p)
      inv -= p;
    mpqs_2A_all_inv[i] = inv;
  }
  mpqs_FB[2 * mpqs_nFB] = mpqs_sievelen;
  return 1;
}

/*******************************************************/
static int mpqs_next_pol()
/*******************************************************/
{ long i, ind, j;
  long long add, bi, prod_other;
  ushort p, s1, s2, inv, *fb = mpqs_FB, *fbs = mpqs_FB_start;
  short sh;
  ulong bb, cc, c1, c2;
  uchar mask;
  double d;
  ushort aa;
  ulong al, *ul;

  mpqs_B_index++;
  if (mpqs_B_index >= 1 << (mpqs_nAdiv - 1)) {
/*zeitA(7);*/
    mpqs_A_index++;
    if (mpqs_A_index >= mpqs_nA)
      return 0;
    mask = mpqs_A_mask[mpqs_A_index];
    for (i = 0; i < mpqs_nAdiv_total; i++)
      if (mask & (1 << i))
        mpqs_Adiv_active[i] = 1;
      else
        mpqs_Adiv_active[i] = 0;
    j = 0;
    for (i = 0; i < mpqs_nAdiv_total; i++)
      if (mpqs_Adiv_active[i]) {
        mpqs_Adiv[j] = mpqs_Adiv_all[i];
        mpqs_Adiv_sqrt[j] = mpqs_Adiv_all_sqrt[i];
        j++;
      }
    mpqs_A = 1;
    for (i = 0; i < mpqs_nAdiv; i++)
      mpqs_A *= (long long) (mpqs_Adiv[i]);
/* compute zeros */
    d = sqrt(mpqs_kN_dbl) / (double) mpqs_A;
    if (d < (double) mpqs_disp) {
      mpqs_zero1 = mpqs_disp - (ushort) d;
      mpqs_zero2 = mpqs_disp + (ushort) d;
    } else {
      mpqs_zero1 = mpqs_sievelen;
      mpqs_zero2 = mpqs_sievelen;
    }
/* compute B_i's */
    for (i = 0; i < mpqs_nAdiv; i++) {
      p = mpqs_Adiv[i];
      bi = mpqs_A / (long long) (p);
      inv = (ushort) (bi % (long long) p);      /* bi>0 */
      inv = mpqs_invert(inv, p);
      bi *= (long long) inv;
      bi %= mpqs_A;
      bi *= (long long) mpqs_Adiv_sqrt[i];
      bi %= mpqs_A;
      mpqs_Bi[i] = bi;
    }
    mpqs_B = 0;
    for (i = 0; i < mpqs_nAdiv; i++)
      mpqs_B += mpqs_Bi[i];
    prod_other = 1;
    for (i = 0; i < mpqs_nAdiv_total; i++)
      if (!mpqs_Adiv_active[i])
        prod_other *= (long long) (mpqs_Adiv_all[i]);
    for (i = 1; i < mpqs_nFB; i++) {
      p = fb[2 * i];
#if 0
      bb = (ulong) (prod_other % (long long) p);
      bb *= (ulong) (mpqs_2A_all_inv[i]);
      bb %= (ulong) p;
#else
      bb =
        (ulong) (((long long) (mpqs_2A_all_inv[i]) * prod_other) % (ulong) p);
#endif
      for (j = 0; j < mpqs_nAdiv; j++) {
#if 1
        cc = bb * (ulong) (mpqs_Bi[j] % (long long) p);
        cc %= (ulong) p;
#else
        cc = (ulong) (((long long) bb * mpqs_Bi[j]) % (ulong) p);
#endif
        mpqs_SI_add[j][i] = (ushort) cc;
      }
      if (bb & 1)
        bb += p;
      bb >>= 1;                 /* now bb=1/A mod p */
      cc = p - (ulong) (mpqs_B % (long long) p);        /* mpqs_B>0 */
      c1 = cc + fb[2 * i + 1];
      c1 *= bb;
      c1 += mpqs_disp;
      c1 %= (ulong) p;
      c2 = cc + (p - fb[2 * i + 1]);
      c2 *= bb;
      c2 += mpqs_disp;
      c2 %= (ulong) p;
#if 0
      if (c1 < c2) {
        fbs[2 * i] = (ushort) c1;
        fbs[2 * i + 1] = (ushort) c2;
      } else {
        fbs[2 * i] = (ushort) c2;
        fbs[2 * i + 1] = (ushort) c1;
      }
#else
      fbs[2 * i] = (ushort) c1;
      fbs[2 * i + 1] = (ushort) c2;
#endif
    }

    mpqs_2B = 2 * mpqs_B;
    mpqs_A_inv_64 = mpqs_inv_64(mpqs_A);
#if 0
    mpz_set_sll(mpqs_help, mpqs_B);
    mpz_mul(mpqs_help, mpqs_help, mpqs_help);
    mpz_sub(mpqs_help, mpqs_help, mpqs_kN);
    mpz_set_sll(mpqs_dummy, mpqs_A);
    mpz_fdiv_qr(mpqs_help, mpqs_dummy, mpqs_help, mpqs_dummy);
    if (mpz_sgn(mpqs_dummy)) {
      mpz_out_str(stdout, 10, mpqs_help);
      printf("\n");
      mpz_out_str(stdout, 10, mpqs_dummy);
      printf("\n");
      Schlendrian("mpqs_next_pol\n");
    }
    mpz_get_sll(&mpqs_C, mpqs_help);
#else
    mpqs_C = mpqs_B * mpqs_B;
    mpqs_C -= mpqs_kN_64;
    mpqs_C *= mpqs_A_inv_64;
#endif

    for (i = 0; i < mpqs_nAdiv_total; i++)
      if (!mpqs_Adiv_active[i]) {
        p = mpqs_Adiv_all[i];
        bb = (ulong) (mpqs_A % (long long) p);
        bb = mpqs_invert(bb, p);
        bb <<= 1;
        if (bb >= p)
          bb -= p;
        for (j = 0; j < mpqs_nAdiv; j++) {
          cc = bb * (ulong) (mpqs_Bi[j] % (long long) p);
          cc %= (ulong) p;
          mpqs_Adiv_SI_add[j][i] = (ushort) cc;
        }
        if (bb & 1)
          bb += p;
        bb >>= 1;               /* now bb=1/A mod p */
        cc = p - (ulong) (mpqs_B % (long long) p);      /* mpqs_B>0 */
        c1 = cc + mpqs_Adiv_all_sqrt[i];
        c1 *= bb;
        c1 += mpqs_disp;
        c1 %= (ulong) p;
        c2 = cc + (p - mpqs_Adiv_all_sqrt[i]);
        c2 *= bb;
        c2 += mpqs_disp;
        c2 %= (ulong) p;
        if (c1 < c2) {
          mpqs_Adiv_start1[i] = (ushort) c1;
          mpqs_Adiv_start2[i] = (ushort) c2;
        } else {
          mpqs_Adiv_start1[i] = (ushort) c2;
          mpqs_Adiv_start2[i] = (ushort) c1;
        }
      } else {
        p = mpqs_Adiv_all[i];
        sh = (short) (mpqs_2B % (long long) p);
        if (sh > 0)
          bb = (ulong) sh;
        else
          bb = (ulong) ((short) p + sh);
        bb = (ulong) mpqs_invert(bb, p);
        sh = (short) (mpqs_C % (long long) p);
        sh = -sh;
        if (sh > 0)
          cc = (ulong) sh;
        else
          cc = (ulong) ((short) p + sh);
        cc *= bb;
        cc %= (ulong) p;        /* now p|2*B*cc+C */
        cc += mpqs_disp;
        cc %= (ulong) p;
        mpqs_Adiv_start1[i] = (ushort) cc;
      }
    mpqs_B_index = 0;
/*zeitB(7);*/
    return 1;
  }
/*zeitA(7);*/
  ind = mpqs_B_index;
  i = 0;
  while (1) {
    if (ind & 1)
      break;
    i++;
    ind >>= 1;
  }
  ind >>= 1;
  ind++;
  add = 2 * mpqs_Bi[i];
  if (ind & 1) {
    mpqs_B -= add;
    for (j = 1; j < mpqs_nFB; j++) {
#if 0
      p = fb[2 * j];
      s1 = fbs[2 * j];
      s2 = fbs[2 * j + 1];
      s1 += mpqs_SI_add[i][j];
      if (s1 >= p)
        s1 -= p;
      s2 += mpqs_SI_add[i][j];
      if (s2 >= p)
        s2 -= p;
/*      if (s1<s2) {*/
      fbs[2 * j] = s1;
      fbs[2 * j + 1] = s2;
/*      } else {
        fbs[2*j]=s2; fbs[2*j+1]=s1;
      }*/
#else
      p = fb[2 * j];            /*s1=fbs[2*j]; s2=fbs[2*j+1]; */
      aa = mpqs_SI_add[i][j];
      al = (((ulong) aa) << 16) + (ulong) aa;
      ul = (ulong *) (fbs + 2 * j);
      *ul += al;
/*      fbs[2*j]+=aa; fbs[2*j+1]+=aa;*/
      if (fbs[2 * j] >= p)
        fbs[2 * j] -= p;
      if (fbs[2 * j + 1] >= p)
        fbs[2 * j + 1] -= p;
/*      s1+=aa; if (s1>=p) s1-=p;
      s2+=aa; if (s2>=p) s2-=p;
      fbs[2*j]=s1; fbs[2*j+1]=s2;*/
#endif
    }
  } else {
    mpqs_B += add;
    for (j = 1; j < mpqs_nFB; j++) {
#if 0
      p = fb[2 * j];
      s1 = fbs[2 * j];
      s2 = fbs[2 * j + 1];
      s1 += (p - mpqs_SI_add[i][j]);
      if (s1 >= p)
        s1 -= p;
      s2 += (p - mpqs_SI_add[i][j]);
      if (s2 >= p)
        s2 -= p;
/*      if (s1<s2) {*/
      fbs[2 * j] = s1;
      fbs[2 * j + 1] = s2;
/*      } else {
        fbs[2*j]=s2; fbs[2*j+1]=s1;
      }*/
#else
      p = fb[2 * j];            /*s1=fbs[2*j]; s2=fbs[2*j+1]; */
      aa = p - mpqs_SI_add[i][j];
      al = (((ulong) aa) << 16) + (ulong) aa;
      ul = (ulong *) (fbs + 2 * j);
      *ul += al;
/*      fbs[2*j]+=aa; fbs[2*j+1]+=aa;*/
      if (fbs[2 * j] >= p)
        fbs[2 * j] -= p;
      if (fbs[2 * j + 1] >= p)
        fbs[2 * j + 1] -= p;
/*      s1+=aa; if (s1>=p) s1-=p;
      s2+=aa; if (s2>=p) s2-=p;
      fbs[2*j]=s1; fbs[2*j+1]=s2;*/
#endif
    }
  }
/*zeitB(7);*/
  mpqs_2B = 2 * mpqs_B;
#if 0
  mpz_set_sll(mpqs_help, mpqs_B);
  mpz_mul(mpqs_help, mpqs_help, mpqs_help);
  mpz_sub(mpqs_help, mpqs_help, mpqs_kN);
  mpz_set_sll(mpqs_dummy, mpqs_A);
#if 0
  mpz_fdiv_qr(mpqs_help, mpqs_dummy, mpqs_help, mpqs_dummy);
  if (mpz_sgn(mpqs_dummy))
    Schlendrian("mpqs_next_pol\n");
#else
  mpz_divexact(mpqs_help, mpqs_help, mpqs_dummy);
#endif
  mpz_get_sll(&mpqs_C, mpqs_help);
#else
  mpqs_C = mpqs_B * mpqs_B;
  mpqs_C -= mpqs_kN_64;
  mpqs_C *= mpqs_A_inv_64;
#endif
/*printf("pol: %Ld %Ld %Ld \n",mpqs_A,mpqs_B,mpqs_C);*/
  for (j = 0; j < mpqs_nAdiv_total; j++)
    if (!mpqs_Adiv_active[j]) {
      p = mpqs_Adiv_all[j];
      s1 = mpqs_Adiv_start1[j];
      s2 = mpqs_Adiv_start2[j];
      if (ind & 1) {
        s1 += mpqs_Adiv_SI_add[i][j];
        if (s1 >= p)
          s1 -= p;
        s2 += mpqs_Adiv_SI_add[i][j];
        if (s2 >= p)
          s2 -= p;
      } else {
        s1 += (p - mpqs_Adiv_SI_add[i][j]);
        if (s1 >= p)
          s1 -= p;
        s2 += (p - mpqs_Adiv_SI_add[i][j]);
        if (s2 >= p)
          s2 -= p;
      }
      mpqs_Adiv_start1[j] = s1;
      mpqs_Adiv_start2[j] = s2;
    } else {
      p = mpqs_Adiv_all[i];
      sh = (short) (mpqs_2B % (long long) p);
      if (sh > 0)
        bb = (ulong) sh;
      else
        bb = (ulong) ((short) p + sh);
      bb = (ulong) mpqs_invert(bb, p);
      sh = (short) (mpqs_C % (long long) p);
      sh = -sh;
      if (sh > 0)
        cc = (ulong) sh;
      else
        cc = (ulong) ((short) p + sh);
      cc *= bb;
      cc %= (ulong) p;          /* now p|2*B*cc+C */
      cc += mpqs_disp;
      cc %= (ulong) p;
      mpqs_Adiv_start1[i] = (ushort) cc;
    }
/*zeitB(7);*/
  return 1;
}

/*******************************************************/
static void mpqs_sieve()
/*******************************************************/
{ uchar *sv = mpqs_sievearray, *fblog = mpqs_FB_log, *svend;
  ushort *fb = mpqs_FB, *fbs = mpqs_FB_start;
  ushort p, s1, s2, end, pb;
  long i;
  uchar lo;
  ulong *ulsv, *ulsvend, mask;

  ulsv = (ulong *) mpqs_sievearray;
  ulsvend = ulsv + mpqs_sievelen / 4;
  if (mpqs_B & 1) {
    mask = (ulong) (mpqs_FB_log[0]);
    mask *= 0x00040004;
  } else {
    mask = (ulong) (mpqs_FB_log[0]);
    mask *= 0x04000400;
  }
  while (ulsv < ulsvend) {
    *ulsv++ = mask;
    *ulsv++ = mask;
    *ulsv++ = mask;
    *ulsv++ = mask;
  }
/*printf("%lu %lu \n",(ulong)(mpqs_sievearray[0]),(ulong)(mpqs_sievearray[1]));*/
/*  mask*=0x00000101;
  if (mpqs_zero1<mpqs_sievelen) {
    s1=mpqs_zero1/4; if (s1<128) s1=0; else s1-=128;
    ulsv=(ulong *)mpqs_sievearray; ulsv+=s1; ulsvend=ulsv+256;
    while (ulsv<ulsvend) { *ulsv++=mask; *ulsv++=mask; *ulsv++=mask; *ulsv++=mask; }
  }
  if (mpqs_zero2<mpqs_sievelen) {
    s1=mpqs_zero2/4; if (s1>(mpqs_sievelen/4)-128) s1=(mpqs_sievelen/4)-256; else s1-=128;
    ulsv=(ulong *)mpqs_sievearray; ulsv+=s1; ulsvend=ulsv+256;
    while (ulsv<ulsvend) { *ulsv++=mask; *ulsv++=mask; *ulsv++=mask; *ulsv++=mask; }
  }*/
  pb = mpqs_sievelen >> 2;
  i = mpqs_sievebegin;
  fb += 2 * i;
  fbs += 2 * i;

#if 1
  asm_sieve();
#else
  for (;;) {
    p = *fb;
    if (p > pb)
      break;
    lo = fblog[i];
    sv = mpqs_sievearray;
    fb += 2;
    s1 = *fbs++;
    s2 = *fbs++;
    svend = sv + mpqs_sievelen - 4 * p;
    while (sv < svend) {
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
    }
    svend += (p + p);
    if (sv < svend) {
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
    }
    svend += p;
    if (sv < svend) {
      sv[s1] += lo;
      sv[s2] += lo;
      sv += p;
    }
    svend += p;
    if (sv + s1 < svend)
      sv[s1] += lo;
    if (sv + s2 < svend)
      sv[s2] += lo;
    i++;
  }
#endif
  sv = mpqs_sievearray;
#ifdef MPQS_ZEIT
  zeita(6);
#endif
  for (i = 0; i < mpqs_nAdiv_total; i++)
    if (!mpqs_Adiv_active[i]) {
      p = mpqs_Adiv_all[i];
      lo = mpqs_Adiv_log[i];
      s1 = mpqs_Adiv_start1[i];
      s2 = mpqs_Adiv_start2[i];
      while (s1 < mpqs_sievelen) {
        sv[s1] += lo;
        s1 += p;
      }
      while (s2 < mpqs_sievelen) {
        sv[s2] += lo;
        s2 += p;
      }
    } else {
      p = mpqs_Adiv_all[i];
      lo = mpqs_Adiv_log[i];
      s1 = mpqs_Adiv_start1[i];
      while (s1 < mpqs_sievelen) {
        sv[s1] += lo;
        s1 += p;
      }
    }
#ifdef MPQS_ZEIT
  zeitb(6);
#endif
}

/*******************************************************/
static ushort mpqs_evaluate()
/*******************************************************/
{ ulong *ulsv, *ulsvend, h;
  uchar *sv;
  ushort **rels, buffer[256];
  long i, nmaxsurv;

  mpqs_nsurvivors = 0;
  nmaxsurv = MPQS_MAX_NRELS - mpqs_nrels;
  if (MPQS_MAX_NPRELS - mpqs_nprels < nmaxsurv)
    nmaxsurv = MPQS_MAX_NPRELS - mpqs_nprels;
  if (MPQS_MAX_NCRELS - mpqs_ncrels < nmaxsurv)
    nmaxsurv = MPQS_MAX_NCRELS - mpqs_ncrels;
  if (nmaxsurv < MPQS_MIN_RELBUFFER)
    return 0;
  if (nmaxsurv > 255)
    nmaxsurv = 255;             /* 1 weniger als buffer-Lnge !!! */
  ulsv = (ulong *) mpqs_sievearray;
  ulsvend = ulsv + (mpqs_sievelen >> 2);
  rels = mpqs_rel_buffer;
#if 0
  while (ulsv < ulsvend) {
    h = *ulsv;
    if (h & 0x80808080) {
      sv = (uchar *) ulsv;
      for (i = 0; i < 4; i++)
        if (*sv++ & 0x80)
          buffer[mpqs_nsurvivors++] = (ushort) (sv - mpqs_sievearray - 1);
    }
    *ulsv++ = 0;
    h = *ulsv;
    if (h & 0x80808080) {
      sv = (uchar *) ulsv;
      for (i = 0; i < 4; i++)
        if (*sv++ & 0x80)
          buffer[mpqs_nsurvivors++] = (ushort) (sv - mpqs_sievearray - 1);
      if (mpqs_nsurvivors > nmaxsurv - 8) {
        while (ulsv < ulsvend)
          *ulsv++ = 0;
        break;
      }
    }
    *ulsv++ = 0;
  }
#else
  mpqs_nsurvivors = asm_evaluate(ulsv, ulsvend, buffer, nmaxsurv);      /*printf("%ld ",mpqs_nsurvivors); */
#ifdef MPQS_STAT
  stat_asm_eval++;
#endif
#endif

  for (i = 0; i < mpqs_nsurvivors; i++) {
    mpqs_sievearray[buffer[i]] = (uchar) (i + 1);
    rels[i][0] = buffer[i];
  }
  return mpqs_nsurvivors;
}

/*******************************************************/
/* change hash_table size to 256*15 ??? */
static inline ushort mpqs_hash_lp(ulong lp)
/*******************************************************/
{ long j;
  ulong lp1;
  ushort lphash, nh;

  lphash = (ushort) (lp & 0x00ff);
  lphash >>= 1;
  lp1 = lp >> 8;
  nh = mpqs_hash_table[lphash][0];
  if (nh >= 15) {               /*printf("."); */
    return MPQS_HASH_OVERFLOW;
  }
  for (j = 1; j <= nh; j++)
    if (mpqs_hash_table[lphash][j] == lp1)
      break;
  if (j > nh) {                 /* new lp */
    mpqs_hash_table[lphash][j] = lp1;
    mpqs_hash_table[lphash][0]++;
    mpqs_nlp++;
    mpqs_hash_index[lphash][j] = mpqs_nprels;
    mpqs_lp[mpqs_nprels] = lp;
  }
  return mpqs_hash_index[lphash][j];
}

/*******************************************************/
static inline int mpqs_hash_rel(long long axb)
/*******************************************************/
{ long j;
  ulong hash, hash2, nh;

  hash = (ulong) (axb & 0xffffff);
  hash2 = (ushort) (hash >> 8);
  hash &= 0xff;
  nh = mpqs_rel_hash_table[hash][0];
  if (nh >= 15)
    return 1;
  for (j = 1; j <= nh; j++)
    if (mpqs_rel_hash_table[hash][j] == hash2)
      return 1;
  mpqs_rel_hash_table[hash][nh + 1] = hash2;
  mpqs_rel_hash_table[hash][0]++;
  return 0;
}

/*******************************************************/
static void mpqs_decompose()
/*******************************************************/
{ long i, j;
  ushort *fb = mpqs_FB, *fbs = mpqs_FB_start;
  uchar *sv, *svend, i1, i2;
  ushort ind, p, s1, s2, nr, d, lpi, r, ii;
  ushort **rels;
  long long axb, *llp;
  ull qx, pr, prasm;
  ushort d2, ii2, minus;
  double dbl_qx, dx;
  short x;
  ulong *ultarg, *ulsrc, lp, ulqx;
  ulong inv, h, hi, ls1, ls2;
  ulong ax, ay, az, at, asmh;
  ulong rr, jj;

#ifdef MPQS_ZEIT
  zeitA(11);
#endif
  rels = mpqs_rel_buffer - 1;
  for (i = 1; i <= mpqs_nsurvivors; i++)
    rels[i][4] = 0;
  i = mpqs_td_begin;
  fb += 2 * i;
  fbs += 2 * i;
  for (; i < mpqs_nFB; i++) {
    p = *fb;
    fb += 2;
    sv = mpqs_sievearray;
    svend = sv + mpqs_sievelen - 2 * p;
    s1 = *fbs++;
    s2 = *fbs++;
    while (sv < svend) {
      i1 = sv[s1];
      i2 = sv[s2];
      sv += p;
      if (i1 || i2) {
        if (i1) {
          nr = rels[i1][4];
          rels[i1][nr + 5] = mpqs_nFBk_1 + i;
          rels[i1][4]++;
        }
        if (i2) {
          nr = rels[i2][4];
          rels[i2][nr + 5] = mpqs_nFBk_1 + i;
          rels[i2][4]++;
        }
      }
      i1 = sv[s1];
      i2 = sv[s2];
      sv += p;
      if (i1 || i2) {
        if (i1) {
          nr = rels[i1][4];
          rels[i1][nr + 5] = mpqs_nFBk_1 + i;
          rels[i1][4]++;
        }
        if (i2) {
          nr = rels[i2][4];
          rels[i2][nr + 5] = mpqs_nFBk_1 + i;
          rels[i2][4]++;
        }
      }
    }
    svend += p;
    while (sv < svend) {
      i1 = sv[s1];
      i2 = sv[s2];
      sv += p;
      if (i1 || i2) {
        if (i1) {
          nr = rels[i1][4];
          rels[i1][nr + 5] = mpqs_nFBk_1 + i;
          rels[i1][4]++;
        }
        if (i2) {
          nr = rels[i2][4];
          rels[i2][nr + 5] = mpqs_nFBk_1 + i;
          rels[i2][4]++;
        }
      }
    }
    svend += p;
    if (sv + s1 < svend) {
      i1 = sv[s1];
      if (i1) {
        nr = rels[i1][4];
        rels[i1][nr + 5] = mpqs_nFBk_1 + i;
        rels[i1][4]++;
      }
    }
    if (sv + s2 < svend) {
      i2 = sv[s2];
      if (i2) {
        nr = rels[i2][4];
        rels[i2][nr + 5] = mpqs_nFBk_1 + i;
        rels[i2][4]++;
      }
    }
  }
#ifdef MPQS_ZEIT
  zeitB(11);
#endif
  for (i = 1; i <= mpqs_nsurvivors; i++) {
    nr = rels[i][4];
    ind = rels[i][0];
    x = (short) (ind) - mpqs_disp;

    axb = mpqs_A * (long long) x + mpqs_B;
    if (axb < 0)
      axb = -axb;
    if (mpqs_hash_rel(axb))
      goto next;

    qx = mpqs_A * (ull) x + mpqs_2B;
    qx *= (ull) x;
    qx += (ull) mpqs_C;

/* das folgende besser durch ungefhre Berechnung der Nullstellen
   und Vergleiche ersetzen ?? */
    minus = 0;
    dx = (double) x;
    dbl_qx = (double) mpqs_A *dx + (double) mpqs_2B;

    dbl_qx *= dx;
    dx = (double) mpqs_C;
    if (dx > 0.)
      dx -= 18446744073709551616.;
    dbl_qx += dx;
    if (fabs(dbl_qx) >= 18446744073709551616.) {        /*printf(";"); */
      goto next;
    }
    if (dbl_qx < 0.) {
      minus = 1;
      qx = -qx;
    }
#if 1
#ifdef MPQS_ZEIT
    zeitA(12);
#endif
#ifdef MPQS_STAT
    stat_asm_td++;
#endif
    if (asm_td(rels[i], minus, &qx, &ulqx)) {
#ifdef MPQS_ZEIT
      zeitB(12);
#endif
      goto next;
    }
/*
if (nr>=40) {
complain("%Lu %ld, %ld, ",qx,ulqx,nr);
}
*/
/*printf(",");*/
/*if (nr>27) {
#ifdef MPQS_ZEIT
 zeitB(12);
#endif
  goto next;
}*/
/*printf("%lu ",ulqx);*/
/*      for (j=0; j<mpqs_nFBk; j++) {
        p=mpqs_FBk[j];
        if (ulqx%(ulong)p==0) {
          complain("%Lu %lu %lu ",qx,ulqx,p);
        }
      }*/
/*      for (j=0; j<mpqs_nFBk; j++) {
        p=mpqs_FBk[j];
        if (ulqx%(ulong)p==0) {
          complain("%Lu %lu %lu ",qx,ulqx,p);
        }
      }*/

#else
    for (j = 1; j < mpqs_td_begin; j++) {
      p = mpqs_FB[2 * j];
      inv = mpqs_FB_inv[j];
      ls1 = ind + (p - mpqs_FB_start[2 * j]);
      ls2 = ind + (p - mpqs_FB_start[2 * j + 1]);
      ls1 *= inv;
      ls2 *= inv;
      ls1 &= 0xffff0000;
      ls2 &= 0xffff0000;
      if (!ls1 || !ls2) {
        rels[i][5 + nr] = mpqs_nFBk_1 + j;
        nr++;                   /*printf("%d.",p); */
      }
    }

    pr = 1;
    for (j = 0; j < nr; j++)
      pr *= (ull) (mpqs_FB[2 * (rels[i][5 + j] - mpqs_nFBk_1)]);
    rels[i][4] = nr;
    if (minus) {
      rels[i][5 + rels[i][4]] = 0;
      rels[i][4]++;
    }
    while (!(qx & 1)) {
      if (rels[i][4] < MPQS_TD_MAX_NDIV) {
        rels[i][5 + rels[i][4]] = mpqs_nFBk_1;
        rels[i][4]++;
        qx >>= 1;
      } else {
#ifdef MPQS_ZEIT
        zeitB(12);
#endif
        goto next;
      }
    }
    if (pr < 4294967296) {
      if (qx > (pr << 32)) {
/*        printf(",");*/
#ifdef MPQS_ZEIT
        zeitB(12);
#endif
        goto next;
      }
    }
    ax = (ulong) qx;
    ay = (ulong) pr;            /*az=0; */
    inv = (ulong) mpqs_256_inv_table[(ay & 0xff) >> 1];
    at = ay * inv;
    at &= 0x0000ff00;
    at *= inv;
    inv -= at;
    at = ay * inv;
    at &= 0xffff0000;
    at *= inv;
    inv -= at;
/*    if (ay*inv!=1) complain("dec.neu8");*/
    az = ax * inv;
/*if (qx%pr) complain("a");
if (qx/pr!=az) complain("b");*/
    ulqx = az;

    for (j = 0; j < nr; j++) {
      ii = rels[i][5 + j];
      p = mpqs_FB[2 * (ii - mpqs_nFBk_1)];
      while (ulqx % (ulong) p == 0) {
        if (rels[i][4] < MPQS_TD_MAX_NDIV) {
          rels[i][5 + rels[i][4]] = ii;
          rels[i][4]++;
          ulqx /= (ulong) p;
        } else {
#ifdef MPQS_ZEIT
          zeitB(12);
#endif
          goto next;
        }
      }
    }

    if (ulqx > 1)
      for (j = 0; j < mpqs_nFBk; j++) {
        p = mpqs_FBk[j];
        while (ulqx % (ulong) p == 0) {
          if (rels[i][4] < MPQS_TD_MAX_NDIV) {
            rels[i][5 + rels[i][4]] = 1 + j;
            rels[i][4]++;
            ulqx /= (ulong) p;
          } else {
#ifdef MPQS_ZEIT
            zeitB(12);
#endif
            goto next;
          }
        }
        if (ulqx == 1)
          break;
      }
    if (ulqx > 1)
      for (j = 0; j < mpqs_nAdiv_total; j++) {
        p = mpqs_Adiv_all[j];
        while (ulqx % (ulong) p == 0) {
          if (rels[i][4] < MPQS_TD_MAX_NDIV) {
            rels[i][5 + rels[i][4]] = 1 + mpqs_nFB + mpqs_nFBk + j;
            rels[i][4]++;
            ulqx /= (ulong) p;
          } else {
#ifdef MPQS_ZEIT
            zeitB(12);
#endif
            goto next;
          }
        }
        if (ulqx == 1)
          break;
      }
#endif

    if (rels[i][4] < MPQS_TD_MAX_NDIV - mpqs_nAdiv) {
      nr = rels[i][4];
      ind = 0;
      for (j = 0; j < mpqs_nAdiv_total; j++)
        if (mpqs_Adiv_active[j]) {
          rels[i][5 + nr + ind] = 1 + mpqs_nFB + mpqs_nFBk + j;
          ind++;
        }
      rels[i][4] += mpqs_nAdiv;
    } else {
#ifdef MPQS_ZEIT
      zeitB(12);
#endif
      goto next;
    }
#ifdef MPQS_ZEIT
    zeitB(12);
#endif
    if (ulqx == 1) {
      llp = (long long *) (mpqs_relations[mpqs_nrels]);
      *llp = axb;
      nr = rels[i][4];
      for (j = 0; j <= nr; j++)
        mpqs_relations[mpqs_nrels][j + 4] = rels[i][j + 4];
      mpqs_nrels++;
      goto next;
    }
    if (ulqx < 1048576) {
      if (ulqx <= mpqs_pmax) {
        printf("%d %Ld %lu ", x, qx, ulqx);
        complain("dec.2\n");
      }
      lp = ulqx;
      lpi = mpqs_hash_lp(lp);
      if (lpi == MPQS_HASH_OVERFLOW)
        goto next;
      if (lpi < mpqs_nprels) {
        if (mpqs_lp[lpi] != lp)
          Schlendrian("lp");
        llp = (long long *) (mpqs_comb_rels[mpqs_ncrels]);
        *llp = axb;
        llp = (long long *) (mpqs_part_rels[lpi]);
        axb = *llp;
        llp = (long long *) (mpqs_comb_rels[mpqs_ncrels]);
        llp++;
        *llp = axb;
        nr = rels[i][4];
        for (j = 0; j < nr; j++)
          mpqs_comb_rels[mpqs_ncrels][j + 9] = rels[i][j + 5];
        for (j = 0; j < mpqs_part_rels[lpi][4]; j++)
          mpqs_comb_rels[mpqs_ncrels][j + 9 + nr] =
            mpqs_part_rels[lpi][j + 5];
        mpqs_comb_rels[mpqs_ncrels][8] = nr + 256 * mpqs_part_rels[lpi][4];
        mpqs_ncrels++;
      } else {
        llp = (long long *) (mpqs_part_rels[mpqs_nprels]);
        *llp = axb;
        nr = rels[i][4];
        for (j = 0; j <= nr; j++)
          mpqs_part_rels[mpqs_nprels][j + 4] = rels[i][j + 4];
        mpqs_nprels++;
      }
    }                           /*else printf("<"); */
  next:
    ;                           /* CJM: Null statement to hush compiler warning. */
  }
  if (mpqs_nsp < mpqs_nrels + mpqs_ncrels)
    mpqs_excess = mpqs_nrels + mpqs_ncrels - mpqs_nsp;
  else
    mpqs_excess = 0;
}

/*
order of the primes:
-1, primes in mpqs_FBk, primes in mpqs_FB, primes in mpqs_Adiv_total
*/

/*
static ulong mpqs_gauss_mask0[]  =
{
  0x80000000UL, 0x40000000UL, 0x20000000UL, 0x10000000UL,
  0x08000000UL, 0x04000000UL, 0x02000000UL, 0x01000000UL,
  0x00800000UL, 0x00400000UL, 0x00200000UL, 0x00100000UL,
  0x00080000UL, 0x00040000UL, 0x00020000UL, 0x00010000UL,
  0x00008000UL, 0x00004000UL, 0x00002000UL, 0x00001000UL,
  0x00000800UL, 0x00000400UL, 0x00000200UL, 0x00000100UL,
  0x00000080UL, 0x00000040UL, 0x00000020UL, 0x00000010UL,
  0x00000008UL, 0x00000004UL, 0x00000002UL, 0x00000001UL
};
*/
static ulong mpqs_gauss_mask[] = {
  0x00000001UL, 0x00000002UL, 0x00000004UL, 0x00000008UL,
  0x00000010UL, 0x00000020UL, 0x00000040UL, 0x00000080UL,
  0x00000100UL, 0x00000200UL, 0x00000400UL, 0x00000800UL,
  0x00001000UL, 0x00002000UL, 0x00004000UL, 0x00008000UL,
  0x00010000UL, 0x00020000UL, 0x00040000UL, 0x00080000UL,
  0x00100000UL, 0x00200000UL, 0x00400000UL, 0x00800000UL,
  0x01000000UL, 0x02000000UL, 0x04000000UL, 0x08000000UL,
  0x10000000UL, 0x20000000UL, 0x40000000UL, 0x80000000UL
};

#define PRUNING
/*******************************************************/
static void mpqs_matrix_init()
/*******************************************************/
{ long i, j, n32;
  ushort nr, ii, pi;
  ulong mask;

#ifdef PRUNING
  ushort mpqs_gauss_wt[300];    /* verbessern !!!! */

  for (i = 0; i < mpqs_nsp; i++)
    mpqs_gauss_wt[i] = 0;
#endif

  mpqs_gauss_n = mpqs_nrels + mpqs_ncrels;
  if (mpqs_gauss_n >= MPQS_GAUSS_MAX)
    return;
  mpqs_gauss_m = mpqs_nsp;
  if (mpqs_gauss_m >= mpqs_gauss_n)
    Schlendrian("gauss: no excess\n");
/* build matrix */
  mpqs_gauss_n32 = (mpqs_gauss_n + 31) / 32;
  for (i = 1; i < mpqs_gauss_m; i++)
    mpqs_gauss_row[i] = mpqs_gauss_row[i - 1] + mpqs_gauss_n32;
  for (i = 0; i < mpqs_gauss_m; i++)
    for (j = 0; j < mpqs_gauss_n32; j++)
      mpqs_gauss_row[i][j] = 0;
  for (i = 0; i < mpqs_nrels; i++) {
    nr = mpqs_relations[i][4];
    mask = mpqs_gauss_mask[i % 32];
    for (j = 0; j < nr; j++) {
      pi = mpqs_relations[i][5 + j];
      mpqs_gauss_row[pi][i / 32] ^= mask;
#ifdef PRUNING
      mpqs_gauss_wt[pi]++;
#endif
    }
  }
  for (i = 0; i < mpqs_ncrels; i++) {
    nr = mpqs_comb_rels[i][8];
    nr = (nr & 255) + (nr >> 8);
    ii = i + mpqs_nrels;
    mask = mpqs_gauss_mask[ii % 32];
    for (j = 0; j < nr; j++) {
      pi = mpqs_comb_rels[i][9 + j];
      mpqs_gauss_row[pi][ii / 32] ^= mask;
#ifdef PRUNING
      mpqs_gauss_wt[pi]++;
#endif
    }
  }
  for (i = 0; i < mpqs_gauss_m; i++)
    mpqs_gauss_c[i] = -1;
  mpqs_gauss_k = 0;
  for (i = 0; i < mpqs_gauss_n; i++)
    mpqs_gauss_d[i] = -1;

#ifdef PRUNING
/* simple pruning */
  {
    ulong mask, k, k32;

#if 1
/*  ulong anz[5];
  for (j=0; j<5; j++) anz[j]=0;
  for (j=0; j<mpqs_gauss_m; j++)
    if (mpqs_gauss_wt[j]<5) anz[mpqs_gauss_wt[j]]++;
  printf("| %lu: ",mpqs_gauss_m);
  for (j=0; j<5; j++) printf("%lu ",anz[j]); printf("| ");
  for (j=0; j<5; j++) stat_gauss_anz[j]+=anz[j];*/

    for (j = 0; j < mpqs_gauss_m; j++)
      if (mpqs_gauss_wt[j] == 1) {      /*printf("!"); */
        k32 = 0;
        while (mpqs_gauss_row[j][k32] == 0)
          k32++;
        for (k = 0; k < 32; k++)
          if (mpqs_gauss_row[j][k32] & mpqs_gauss_mask[k])
            break;
        k += (k32 << 5);
        if (k >= mpqs_gauss_n)
          Schlendrian("pruning\n");

        mpqs_gauss_d[k] = j;
        mpqs_gauss_c[j] = k;
/* No need to clear the k-th column since it will not appear in a solution. */
      }
#else
    for (j = 0; j < mpqs_gauss_m; j++)
      if (mpqs_gauss_wt[j] == 1) {      /*printf("!"); */
        mpqs_gauss_m--;
        for (k = 0; k < mpqs_gauss_n32; k++)
          mpqs_gauss_row[j][k] = mpqs_gauss_row[mpqs_gauss_m][k];
        mpqs_gauss_wt[j] = mpqs_gauss_wt[mpqs_gauss_m];
        break;
      }
#endif
  }
#endif
}

/*******************************************************/
static int mpqs_matrix()
/*******************************************************/
{ long i, j, l, t, k32;
  ulong mask;

#ifdef MPQS_ZEIT
  zeita(8);
#endif
/* solve matrix */
  while (mpqs_gauss_k < mpqs_gauss_n) {
#ifdef PRUNING
    if (mpqs_gauss_d[mpqs_gauss_k] >= 0) {
      mpqs_gauss_k++;
      continue;
    }
#endif
    mask = mpqs_gauss_mask[mpqs_gauss_k % 32];
    k32 = mpqs_gauss_k / 32;
    j = 0;
    while ((j < mpqs_gauss_m) && (mpqs_gauss_c[j] >= 0 ||
                                  ((mpqs_gauss_row[j][k32] & mask) == 0)))
      j++;
    if (j == mpqs_gauss_m) {
      mpqs_gauss_d[mpqs_gauss_k] = -1;
      break;
    }
    mpqs_gauss_d[mpqs_gauss_k] = j;
    mpqs_gauss_c[j] = mpqs_gauss_k;
    for (t = 0; t < j; t++)
      if (mpqs_gauss_row[t][k32] & mask)
        for (l = k32; l < mpqs_gauss_n32; l++)
          mpqs_gauss_row[t][l] ^= mpqs_gauss_row[j][l];
    for (t = j + 1; t < mpqs_gauss_m; t++)
      if (mpqs_gauss_row[t][k32] & mask)
        for (l = k32; l < mpqs_gauss_n32; l++)
          mpqs_gauss_row[t][l] ^= mpqs_gauss_row[j][l];
    mpqs_gauss_k++;
  }
  if (mpqs_gauss_k >= mpqs_gauss_n) {
#ifdef MPQS_ZEIT
    zeitb(8);
#endif
    return 0;
  }
  for (i = 0; i < mpqs_gauss_n; i++)
    mpqs_sol[i] = 0;
  if (mpqs_gauss_d[mpqs_gauss_k] != -1)
    Schlendrian("gauss1\n");
  for (i = 0; i < mpqs_gauss_k; i++)
    if (mpqs_gauss_d[i] != -1)
      if (mpqs_gauss_row[mpqs_gauss_d[i]][k32] & mask)
        mpqs_sol[i] = 1;
  mpqs_sol[mpqs_gauss_k] = 1;
  mpqs_gauss_k++;
#ifdef MPQS_ZEIT
  zeitb(8);
#endif
  return 1;
}

/*******************************************************/
static int mpqs_final()
/*******************************************************/
{ long i, j, k;
  ushort nr, nr1, nr2;
  ulong p, lp, prod1, prod2;
  ull *up;
  int split;

/*test_rels(); test_combs();*/
  if (!mpqs_nfactors) {
    mpqs_nfactors = 1;
    mpz_set(mpqs_factors[0], mpqs_N);
    mpqs_f_status[0] = 0;
  }
  while (mpqs_matrix()) {
#ifdef MPQS_STAT
    stat_mpqs_ntrials++;
#endif
    for (j = 0; j < mpqs_nsp; j++)
      mpqs_exp[j] = 0;
    for (j = 0; j < mpqs_nrels; j++)
      if (mpqs_sol[j]) {
        nr = mpqs_relations[j][4];
        if (j & 1)              /* 'random' distribution of full relations */
          for (k = 0; k < nr; k++)
            mpqs_exp[mpqs_relations[j][5 + k]]--;
        else
          for (k = 0; k < nr; k++)
            mpqs_exp[mpqs_relations[j][5 + k]]++;
      }
    for (j = 0; j < mpqs_ncrels; j++)
      if (mpqs_sol[mpqs_nrels + j]) {
        nr = mpqs_comb_rels[j][8];
        nr1 = nr & 255;
        nr2 = nr >> 8;
        for (k = 0; k < nr1; k++)
          mpqs_exp[mpqs_comb_rels[j][9 + k]]++;
        for (k = 0; k < nr2; k++)
          mpqs_exp[mpqs_comb_rels[j][9 + nr1 + k]]--;
      }
    for (j = 0; j < mpqs_nsp; j++)
      if (mpqs_exp[j] & 1)
        Schlendrian("final.odd\n");
      else
        mpqs_exp[j] >>= 1;

    mpz_set_ui(mpqs_sq1, 1);
    prod1 = 1;
    mpz_set_ui(mpqs_sq2, 1);
    prod2 = 1;
    for (j = 1; j < mpqs_nsp; j++)
      if (mpqs_exp[j]) {
        if (j < 1 + mpqs_nFBk) {
          p = (ulong) (mpqs_FBk[j - 1]);
        } else if (j < 1 + mpqs_nFB + mpqs_nFBk) {
          k = j - mpqs_nFBk - 1;
          p = (ulong) (mpqs_FB[2 * k]);
        } else {
          k = j - mpqs_nFB - mpqs_nFBk - 1;
          p = (ulong) (mpqs_Adiv_all[k]);
        }
        if (mpqs_exp[j] > 0) {
          for (k = 0; k < mpqs_exp[j]; k++) {
            prod1 *= p;
            if (prod1 & 0xfff00000) {   /* p<4096 */
              mpz_mul_ui(mpqs_sq1, mpqs_sq1, prod1);
              mpz_mod(mpqs_sq1, mpqs_sq1, mpqs_N);
#ifdef MPQS_STAT
              stat_final_mulmod++;
#endif
              prod1 = 1;
            }
          }
        } else {
          for (k = 0; k < -mpqs_exp[j]; k++) {
            prod2 *= p;
            if (prod2 & 0xfff00000) {   /* p<4096 */
              mpz_mul_ui(mpqs_sq2, mpqs_sq2, prod2);
              mpz_mod(mpqs_sq2, mpqs_sq2, mpqs_N);
#ifdef MPQS_STAT
              stat_final_mulmod++;
#endif
              prod2 = 1;
            }
          }
        }
      }
    if (prod1 > 1) {
      mpz_mul_ui(mpqs_sq1, mpqs_sq1, prod1);
      mpz_mod(mpqs_sq1, mpqs_sq1, mpqs_N);
    }
    if (prod2 > 1) {
      mpz_mul_ui(mpqs_sq2, mpqs_sq2, prod2);
      mpz_mod(mpqs_sq2, mpqs_sq2, mpqs_N);
    }

    for (j = 0; j < mpqs_nrels; j++)
      if (mpqs_sol[j]) {
        up = (ull *) (mpqs_relations[j]);       /*printf("Q: %Lu  ",*up); */
        mpz_set_ull(mpqs_gmp_help, *up);
        if (j & 1) {
          mpz_mul(mpqs_sq1, mpqs_sq1, mpqs_gmp_help);
          mpz_mod(mpqs_sq1, mpqs_sq1, mpqs_N);
        } else {
          mpz_mul(mpqs_sq2, mpqs_sq2, mpqs_gmp_help);
          mpz_mod(mpqs_sq2, mpqs_sq2, mpqs_N);
        }
      }

    for (j = 0; j < mpqs_ncrels; j++)
      if (mpqs_sol[mpqs_nrels + j]) {
        up = (ull *) (mpqs_comb_rels[j]);       /*printf("Q: %Lu  ",*up); */
        mpz_set_ull(mpqs_gmp_help, *up);
        mpz_mul(mpqs_sq2, mpqs_sq2, mpqs_gmp_help);
        mpz_mod(mpqs_sq2, mpqs_sq2, mpqs_N);
        up++;
        mpz_set_ull(mpqs_gmp_help, *up);
        mpz_mul(mpqs_sq1, mpqs_sq1, mpqs_gmp_help);
        mpz_mod(mpqs_sq1, mpqs_sq1, mpqs_N);
      }

    mpz_add(mpqs_gmp_help, mpqs_sq1, mpqs_sq2);
    mpz_gcd(mpqs_gmp_help, mpqs_gmp_help, mpqs_N);
    if (mpz_cmp_ui(mpqs_gmp_help, 1) == 0)
      continue;
    if (mpz_cmp(mpqs_gmp_help, mpqs_N) == 0)
      continue;

    for (j = 0; j < mpqs_nfactors; j++)
      if (mpqs_f_status[j] == 0) {
        mpz_gcd(mpqs_sq1, mpqs_gmp_help, mpqs_factors[j]);
        if (mpz_cmp_ui(mpqs_sq1, 1) == 0)
          continue;
        if (mpz_cmp(mpqs_sq1, mpqs_factors[j]) == 0)
          continue;
        mpz_set(mpqs_factors[mpqs_nfactors], mpqs_sq1);
        mpz_divexact(mpqs_factors[j], mpqs_factors[j], mpqs_sq1);
        mpqs_f_status[j] = psp(mpqs_factors[j]);
        mpqs_f_status[mpqs_nfactors] = psp(mpqs_factors[mpqs_nfactors]);
        mpqs_nfactors++;
        break;
      }

    split = 1;
    for (j = 0; j < mpqs_nfactors; j++)
      if (mpqs_f_status[j] == 0)
        split = 0;
    if (split)
      return 1;
    if (mpqs_nfactors >= MPQS_MAX_FACTORS)
      Schlendrian("final: too many factors\n");
  }
  return 0;
}

/*******************************************************/
static long mpqs_factor0(mpz_t N, long max_bits, mpz_t ** factors, long retry)
/*******************************************************/
{ long nbits, i, err;
  int ev;

#ifdef MPQS_ZEIT
  zeita(9);
#endif
#ifdef MPQS_STAT
  stat_mpqs_nsieves = 0;
  stat_mpqs_nsurvivors = 0;
  stat_mpqs_ntrials = 0;
  stat_mpqs_ndiv = 0;
#endif
  if (!mpqs_isinit)
    mpqs_init();
#ifdef MPQS_ZEIT
  zeitb(9);
#endif
  nbits = mpz_sizeinbase(N, 2);
  if (nbits > 96) {
    fprintf(stderr, "warning: mpqs called with >96 Bit\n");
    return -2;
  }
#ifdef MPQS_ZEIT
  zeita(1);
#endif
  mpz_set(mpqs_N, N);
  mpqs_choose_multiplier();
  mpqs_choose_parameter(retry);
  mpqs_generate_FB();
  if ((err = mpqs_SI_init()) < 1) {
    printf("%ld ", err);
    fprintf(stderr, "warning: mpqs: error in self-initialization ");
    mpz_out_str(stderr, 10, N);
    printf("\n");
#ifdef MPQS_ZEIT
    zeitb(1);
#endif
    return -3;
  }
#ifdef MPQS_ZEIT
  zeitb(1);
#endif
  while (1) {                   /*printf("%ld,%ld; ",stat_mpqs_nsieves,mpqs_nrels); */
#ifdef MPQS_ZEIT
    zeita(2);
#endif
    if (!mpqs_next_pol()) {
      if (retry) {
        fprintf(stderr, "warning: not enough polynomials in mpqs with ");
        mpz_out_str(stderr, 10, N);
        fprintf(stderr, "\n");
      }
#ifdef MPQS_ZEIT
      zeitb(2);
#endif
      return -4;
    }
#ifdef MPQS_ZEIT
    zeitb(2);
    zeita(3);
#endif
    mpqs_sieve();
#ifdef MPQS_STAT
    stat_mpqs_nsieves++;
#endif
#ifdef MPQS_ZEIT
    zeitb(3);
    zeita(4);
#endif
    ev = mpqs_evaluate();
#ifdef MPQS_STAT
    stat_mpqs_nsurvivors += ev; /*printf("%d %Ld  ",ev,mpqs_C); */
#endif
#ifdef MPQS_ZEIT
    zeitb(4);
#endif
    if (ev) {
#ifdef MPQS_ZEIT
      zeita(5);
#endif
      mpqs_decompose();         /*test_rels(); */
#ifdef MPQS_ZEIT
      zeitb(5);
#endif
      if (mpqs_nsurvivors && (mpqs_excess > MPQS_MIN_EXCESS)) {
#ifdef MPQS_ZEIT
        zeita(8);
#endif
        mpqs_matrix_init();
#ifdef MPQS_ZEIT
        zeitb(8);
        zeita(10);
#endif
        if (mpqs_final()) {
#ifdef MPQS_ZEIT
          zeitb(10);
#endif
#ifdef MPQS_STAT
          printf("Stat: %lu %lu %lu %lu %lu %lu %lu \n", stat_mpqs_nsieves,
                 stat_mpqs_nsurvivors, stat_mpqs_ntrials, stat_mpqs_ndiv,
                 mpqs_nrels, mpqs_nprels, mpqs_ncrels);
#endif
          for (i = 0; i < mpqs_nfactors; i++)
            if (mpz_sizeinbase(mpqs_factors[i], 2) > max_bits)
              return 0;
          *factors = mpqs_factors;
          return mpqs_nfactors;
        }
#ifdef MPQS_ZEIT
        zeitb(10);
#endif
      }
    }
    if (mpqs_nrels > MPQS_MAX_NRELS - MPQS_MIN_RELBUFFER) {
      fprintf(stderr, "warning: too many relations in mpqs\n");
      return -5;
    }
  }
  return -1;
}

/*******************************************************/
long mpqs_factor(mpz_t N, long max_bits, mpz_t ** factors)
/*******************************************************/
{ long err;

  err = mpqs_factor0(N, max_bits, factors, 0);
  if (err == -4)
    err = mpqs_factor0(N, max_bits, factors, 1);
  if (err == -4)
    err = mpqs_factor0(N, max_bits, factors, 3);
  return err;
}
