/**************************************************************/
/* net.c                                                      */
/* Copyright 2004, Chris Monico.                              */
/**************************************************************/
/*  This file is part of GGNFS.
*
*   GGNFS is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   GGNFS is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with GGNFS; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* This file contains various network functions for implementing
   a client/server functionality for GGNFS. The ``security'' stuff
   is pretty ad-hoc. It's just meant to prevent someone from trivially
   misdirecting a client or the server. If it looks like this might
   actually be a problem, I will come back and do it properly.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmp.h>
#include <sys/types.h>
#include <sys/socket.h>


#define BIT(_i) ((0x00000001)<<(_i))

typedef unsigned long hash_t[4];

/*************************************************************/
unsigned long cheese(unsigned long x)
/*************************************************************/
{ static unsigned char lookup[256];
  static int initialized=0;
  int i, j;
  unsigned char c0,c1,c2,c3;

  if (!initialized) {
    for (i=0; i<256; i++) lookup[i] = (((171*i + 72)&0xFF)^0x6C);
    initialized=1;
  }
  c0=x&0x000000FF;       c1=(x&0x0000FF00)>>8;
  c2=(x&0x00FF0000)>>16; c3=(x&0xFF000000)>>24;
  return lookup[c1]^(lookup[c3]<<8)^(lookup[c0]<<16)^(lookup[c2]<<24);
}

/*************************************************************/
void computeHash(hash_t H, unsigned char *data, long dataSize)
/*************************************************************/
{ unsigned char *pData;
  long           pSize, i;
  int            j, r;
  unsigned long  t[4], u;

  /* Make a copy of the (padded) data to be hashed. */  
  pSize = dataSize + ((16 - dataSize%16)%16);
  pData = (unsigned char *)malloc(pSize*sizeof(char));
  memcpy(pData, data, dataSize);
  for (i=dataSize; i<pSize; i++)
    pData[i]=0x00;

  /* Insert a real hash here. This is just a quick & dirty piece
     of garbage for now until I dig up some GPL'd code for SHA-128 or similar. */
  H[0]=H[1]=H[2]=H[3]=0x00000000;
  for (i=0; i<pSize/16; i++) {
    t[0]=t[1]=t[2]=t[3]=0x00000000;
    for (j=0;j<16; j++) 
      t[j/4] ^= ((unsigned long)pData[16*i + j])<<(8*(j&0x03));
    for (r=0; r<32; r++) {
      t[0] ^= (t[1]<<((r+0)%16))^(t[2]<<((r+7)%32))^(t[3]<<((r+13)%16));
      t[1] ^= (t[0]<<((r+2)%16))^(t[2]<<((r+23)%32))^(t[3]<<((r+8)%16));
      t[2] ^= (t[1]<<((r+17)%16))^(t[3]<<((r+3)%32))^(t[4]<<((r+1)%16));
      t[3] ^= (t[0]<<((r+9)%16))^(t[1]<<((r+19)%32))^(t[2]<<((r+29)%16));
      u =    cheese(t[1])^H[2];
      t[1] = cheese(t[2])^H[3];
      t[2] = cheese(t[3])^H[0];
      t[3] = cheese(t[0])^H[1];
      t[0] = u;
    }
    H[0] = cheese(H[2])^t[0];
    H[1] = cheese(H[3])^t[1];
    H[2] = cheese(H[0])^t[2];
    H[3] = cheese(H[1])^t[3];
  }
}

/*************************************************************/
int getDataFromServer(char *data, long maxSize, char *IP)
/*************************************************************/
{

}


/*************************************************************/
int verifyData(mpz_t pubN, mpz_t pubE, mpz_t sig, char *data, long dataSize)
/*************************************************************/
{ mpz_t  tmp1, tmp2;
  hash_t H;
  int    res=0;

  mpz_init(tmp1); mpz_init(tmp2);
  computeHash(H, data, dataSize);

  mpz_pow_mod(tmp1, sig, pubE, pubN);
  mpz_set_ui(tmp2, H[3]); 
  mpz_mul_2exp(tmp2, tmp2, 32); mpz_add_ui(tmp2, tmp2, H[2]);
  mpz_mul_2exp(tmp2, tmp2, 32); mpz_add_ui(tmp2, tmp2, H[1]);
  mpz_mul_2exp(tmp2, tmp2, 32); mpz_add_ui(tmp2, tmp2, H[0]);
  if (mpz_cmp(tmp1, tmp2))
    res = -1;
  mpz_clear(tmp1); mpz_clear(tmp2);
  return res;
}




