#!/bin/bash
# The path where the binaries are:
GGNFS_BIN_PATH=~/Public/src/ggnfs/src

FORCECC=1
SAVEPAIRS=0
CLEANUP=0

LATSIEVER=$GGNFS_BIN_PATH/gnfs-lasieve4I12e
#LATSIEVER=$GGNFS_BIN_PATH/gnfs-lasieve4I13e
#LATSIEVER=$GGNFS_BIN_PATH/gnfs-lasieve4I14e

###################################################################
# FORCECC:
#   To force a count on the number of full relations, set this to 1.
#   The criteria by which GGNFS tries to guess is still pretty lousy,
#   so you may need to do this once in a while for a sanity check.
#   But if you are concerned about how much time is being wasted
#   counting cycles, set it to zero.
# SAVEPAIRS:
#   If this is 1, the script will save all of the siever output in a gzipped file.
# CLEANUP:
#   If this is 1, the script will clean up after itself.
# LATSIEVER:
#   There are three lattice siever binaries to choose from, depending
#   on the size of the number being factored.
#   12e: GNFS numbers upto about 110 digits (SNFS 150).
#   13e: GNFS numbers upto around 130-140 (SNFS 180).
#   14e: Larger.
####################################################################
####################################################################
###### Probably nothing below here should need to be changed. ######
####################################################################
####################################################################
MAKEFB=$GGNFS_BIN_PATH/makefb
GETDEPS=$GGNFS_BIN_PATH/getdeps
MATSOLVE=$GGNFS_BIN_PATH/matsolve
SQRT=$GGNFS_BIN_PATH/sqrt
PLOT=$GGNFS_BIN_PATH/autogplot.sh

DEPFILE=deps
DEPINF=depinf
PROCRELS=rels.bin
LOGFILE="ggnfs.log"

###########################################################
function plotLP {
  if [ ! -x $PLOT ] ; then return ; fi
  echo "0, 0" > .lprels
  echo "0, 0" > .rels
  echo "0, 0" > .fulls
  grep "largePrimes:" $LOGFILE > .tmp.out
  numLines=`wc .tmp.out | awk '{ print $1}'`
  for name in `seq $numLines`
  do
    read line
    line=${line#*largePrimes:}
    lps=`echo $line | awk -F' ' '{print $1}'`
    rels=`echo $line | awk -F': ' '{print $2}'`
     diff=`echo "$lps - $rels" | bc`
     echo "$rels, $diff" >> .lprels
  done <.tmp.out
  XAXIS="Total relations"
  YAXIS=""
  export XAXIS
  export YAXIS
  $PLOT relations.jpg "ExcessLargePrimes" .lprels
#  grep "rels:" $LOGFILE > .tmp.out
#  numLines=`wc .tmp.out | awk '{ print $1}'`
#  for name in `seq $numLines`
#  do
#    read line
#    line=${line#*rels:}
#    rels=`echo $line | awk -F', ' '{print $1}'`
#    fulls=`echo $line | awk -F':' '{print $3}'`
#    echo "$rels, $fulls" >> .fulls
#  done <.tmp.out
#  $PLOT relations.jpg LargePrimes .lprels Relations .rels Fulls .fulls
  rm -f .lprels .rels .fulls .tmp.out
}
######################################################
function checkParams {
  typeset -i BOK=1
  if [ ! -x $MAKEFB ] ; then echo "Cannot find GGNFS program 'makefb'!"; BOK=0 ; fi
  if [ ! -x $GETDEPS ] ; then echo "Cannot find GGNFS program 'getdeps'!"; BOK=0 ; fi
  if [ ! -x $MATSOLVE ] ; then echo "Cannot find GGNFS program 'matsolve'!"; BOK=0 ; fi
  if [ ! -x $SQRT ] ; then echo "Cannot find GGNFS program 'sqrt'!"; BOK=0 ; fi
  if [ $BOK -eq 0 ] ; then
    echo "Did you set GGNFS_BIN_PATH properly in this script?"
    echo "It is currently set to:"
    echo "GGNFS_BIN_PATH=$GGNFS_BIN_PATH"
    exit -1
  fi
  if [ -z "$N" ] ; then echo "Error: 'n' not supplied!"; exit -1 ; fi
  if [ -z "$M" ] ; then echo "Error: 'm' not supplied!"; exit -1 ; fi
  if [ -z "${COEF[0]}" ] ; then echo "Error: polynomial not supplied!"; exit -1 ; fi
  if [ -z "$SKEW" ] ; then echo "Error: 'skew' not supplied!"; exit -1 ; fi
  if [ $RLIM -eq 0 ] ; then echo "Error: 'rlim' not supplied!"; exit -1 ; fi
  if [ $ALIM -eq 0 ] ; then echo "Error: 'alim' not supplied!"; exit -1 ; fi
  if [ $LPBR -eq 0 ] ; then echo "Error: 'lpbr' not supplied!"; exit -1 ; fi
  if [ $LPBA -eq 0 ] ; then echo "Error: 'lpba' not supplied!"; exit -1 ; fi
  if [ $MFBR -eq 0 ] ; then echo "Error: 'mfbr' not supplied!"; exit -1 ; fi
  if [ $MFBA -eq 0 ] ; then echo "Error: 'mfba' not supplied!"; exit -1 ; fi
  if [ -z "$RLAMBDA" ] ; then echo "Error: 'rlambda' not supplied!"; exit -1 ; fi
  if [ -z "$ALAMBDA" ] ; then echo "Error: 'alambda' not supplied!"; exit -1 ; fi
  if [ $QSTEP -eq 0 ] ; then echo "Error: 'qintsize' not supplied!"; exit -1 ; fi
}


######################################################
function makeJobFile {
  # arg0 = filename
  # arg1 = first q value
  # arg2 = q range size
  FNAME=$1

  rm -f $FNAME
  echo "n: $N" >> $FNAME
  echo "m: $M" >> $FNAME
  # The polynomial coefficients:
  i=0;
  while [ -n "${COEF[$i]}" ] ; do 
    echo ${COEF[$i]} ${COEF[${i}+1]} >> $FNAME
    i=$i+2;
  done;
  echo "skew: $SKEW" >> $FNAME
  echo "rlim: $RLIM" >> $FNAME
  echo "alim: $ALIM" >> $FNAME
  echo "lpbr: $LPBR" >> $FNAME
  echo "lpba: $LPBA" >> $FNAME
  echo "mfbr: $MFBR" >> $FNAME
  echo "mfba: $MFBA" >> $FNAME
  echo "rlambda: $RLAMBDA" >> $FNAME 
  echo "alambda: $ALAMBDA" >> $FNAME 
  echo "q0: $2" >> $FNAME
  echo "qintsize: $3" >> $FNAME
}


############### Begin #################
if [ -z "$1" ] ; then
  echo "USAGE: $0 <polynomial file>"
  exit 0
fi;
NAME=$1
NAME=${NAME%.poly}
echo "Working with NAME=$NAME..."
JOBNAME=$NAME.job

#######################################
# Read parameters from the poly file. #
#######################################
# Strip out comments.
grep -v "^#" $NAME.poly > .$NAME.tmp
# Read the parameters.
N=`grep -m 1 "n:" .$NAME.tmp | awk '{ print $2}'`
M=`grep -m 1 "m:" .$NAME.tmp | awk '{ print $2}'`
COEF=(`grep "c.:" .$NAME.tmp`)

typeset -i RLIM=`grep -m 1 "rlim:" .$NAME.tmp | awk '{ print $2}'`
typeset -i ALIM=`grep -m 1 "alim:" .$NAME.tmp | awk '{ print $2}'`
typeset -i LPBR=`grep -m 1 "lpbr:" .$NAME.tmp | awk '{ print $2}'`
typeset -i LPBA=`grep -m 1 "lpba:" .$NAME.tmp | awk '{ print $2}'`
typeset -i MFBR=`grep -m 1 "mfbr:" .$NAME.tmp | awk '{ print $2}'`
typeset -i MFBA=`grep -m 1 "mfba:" .$NAME.tmp | awk '{ print $2}'`
RLAMBDA=`grep -m 1 "rlambda:" .$NAME.tmp | awk '{ print $2}'`
ALAMBDA=`grep -m 1 "alambda:" .$NAME.tmp | awk '{ print $2}'`
SKEW=`grep -m 1 "skew:" .$NAME.tmp | awk '{ print $2}'`
typeset -i Q0=`grep -m 1 "q0:" .$NAME.tmp | awk '{ print $2}'`
typeset -i QSTEP=`grep -m 1 "qintsize:" .$NAME.tmp | awk '{ print $2}'`
if test $Q0 -le $ALIM ; then
  QSTART=$ALIM
else QSTART=$Q0;
fi
rm -f .$NAME.tmp
checkParams

######################################################
# We won't take the liberty of deleting a dependency #
# file in case this script is run by accident.       #
######################################################
if test -e $DEPFILE ; then
  echo "File $DEPFILE already exists. Delete it and re-run for a fresh factorization"
  exit
fi
if test ! -e $NAME.poly ; then
  echo "Error: Polynomial file $NAME.poly does not exist!"
  exit
fi

#########################################
# Should we resume from an earlier run? #
#########################################
typeset -i RESUME=0
if test -e $PROCRELS.0 ; then
  echo "It appears that an earlier attempt was interrupted in progress. Resume? (y/n)"
  read RESPONSE
  if test $RESPONSE = "Y" ; then RESUME=1; fi
  if test $RESPONSE = "y" ; then RESUME=1; fi;
fi;

############################################
# Setup the parameters for sieving ranges. #
############################################
if test $RESUME -ne 1 ; then
  # Clean up any junk leftover from an earlier attempt.
  rm -f $LOGFILE 
  rm -f cols* deps* factor.easy lpindex* rels*  spairs.out spairs.save.gz
  rm -f $NAME.fb *.afb.0
  echo $DD > $LOGFILE
  # Create the factor base.
  $MAKEFB -rl $RLIM -al $ALIM -lpbr $LPBR -lpba $LPBA -2p -of $NAME.fb -if $NAME.poly
  typeset -i Q0=$QSTART
else
  # Get the Q0 value from tmp.job and just restart from there.
  if test -e $JOBNAME ; then
    Q0LINE=`grep q0 $JOBNAME`;
    typeset -i Q0=${Q0LINE:4}
    echo "Found $Q0LINE in $JOBNAME - resuming from b=$Q0"
  else
    echo "File $JOBNAME does not exist. Could not determine a starting q value!"
    echo "Please enter a starting point for the special q: "
    read Q0
    typeset -i Q0=$Q0
  fi;
fi;

#################################################
# Finally, sieve until `getdeps' creates a      #
# depinf file, signalling that sieving is done. #
#################################################
while test ! -e $DEPINF ; do
  echo "*******************************"
  echo "Lattice sieving q-values from q=$Q0 to $Q0+$QSTEP"
  # Create a job file.
  makeJobFile $JOBNAME $Q0 $QSTEP
  typeset -i startTime=`date +"%s"`
  $LATSIEVER -k -o spairs.out -v -a $JOBNAME
  if test ! -e spairs.out ; then
    echo "Some error ocurred and no relations were found! Examing log file."
    exit
  fi
  typeset -i stopTime=`date +"%s"`
  typeset -i totalTime=$stopTime-startTime
  echo "        LatSieveTime: $totalTime" >> $LOGFILE
  
  if [ $FORCECC -eq 1 ] ; then 
    $GETDEPS -fb $NAME.fb -prel $PROCRELS -newrel spairs.out -forceCC
  else
    $GETDEPS -fb $NAME.fb -prel $PROCRELS -newrel spairs.out
  fi
  if test $SAVEPAIRS -eq 1 ; then
    gzip -c spairs.out >> spairs.save.gz
  fi
  rm spairs.out
  Q0=$Q0+QSTEP
  plotLP
done
rm -f $TMPJOBNAME

###############################
# Obviously, the matrix step. #
###############################
echo "Doing matrix step..."
$MATSOLVE
if test ! -e $DEPFILE ; then
  echo "Some error occurred and matsolve did not record dependencies."
  exit -1
fi



#############################################
# Do as many square root jobs as needed to  #
# get the final factorization.              #
#############################################
typeset -i depnum=0
$SQRT $DISC -fb $NAME.fb -deps $DEPFILE -depnum $depnum
typeset -i result=$?
while test $result -ne 0 ; do
  if test $result -eq 1 ; then
    echo -n "Composite factor found. Try another dependency? (y/n)"
    read RESPONSE
    if test $RESPONSE = "N" ; then break; fi
    if test $RESPONSE = "n" ; then break; fi
  fi
  depnum=$depnum+1
  echo "Computing square root for dependency $depnum ..."
  $SQRT $DISC -fb $NAME.fb -deps $DEPFILE -depnum $depnum
  typeset -i result=$?
done;





# Finally, clean up.
if test $CLEANUP -eq 1 ; then
  rm -f cols* deps* factor.easy lpindex* rels*  spairs.out $NAME.fb
  rm -f numFF.txt numrels.txt depinf
fi

######################################################################
# These techniques were borrowed from:
# http://docsrv.caldera.com:8457/en/AdvBashHowto/io-redirection.html
# among other places.
######################################################################

#################################################
# Print some summary data for the factorization #
#################################################
sievetime=0.0 ; grep "SieveTime:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  line=${line#*SieveTime:*}
  sievetime=`echo $sievetime + $line | bc`
done <.tmp.out
sievetime=`echo "scale=5; $sievetime / 3600.0" | bc` # Convert to hours

relproctime=0.0 ; grep "RelProcTime:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  line=${line#*RelProcTime:*}
  relproctime=`echo $relproctime + $line | bc`
done <.tmp.out
relproctime=`echo "scale=5; $relproctime / 3600.0" | bc` # Convert to hours

mattime=0.0 ; grep "BLanczosTime:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  line=${line#*BLanczosTime:*}
  mattime=`echo $mattime + $line | bc`
done <.tmp.out
mattime=`echo "scale=5; $mattime / 3600.0" | bc` # Convert to hours
line=`grep "Initial matrix" $LOGFILE`
initmat=${line#*Initial matrix is*}
line=`grep "Matrix pruned" $LOGFILE`
prunedmat=${line#*Matrix pruned to*}

grep "sqrtTime:" $LOGFILE > .tmp.out
read line <.tmp.out
sqrttime=${line#*sqrtTime:*}
sqrttime=`echo "scale=5; $sqrttime / 3600.0" | bc` # Convert to hours
totaltime=`echo "scale=5; $sievetime+$relproctime+$mattime+$sqrttime" | bc`

grep "rels:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  relsline=${line#*rels:*}
done <.tmp.out
rm -f .tmp.out


echo "Number: $NAME" > summary.txt
echo "Factor base limits: $RLIM/$ALIM" >> summary.txt
echo "Large prime bits: $LPRBITS/$LPABITS" >> summary.txt
typeset -i WIKMB=$Q0+QSTEP
echo "Sieved special-q from $QSTART to $WIKMB" >> summary.txt
echo "Relations: $relsline" >> summary.txt
echo "Initial matrix obtained: $initmat" >> summary.txt
echo "Pruned matrix: $prunedmat" >> summary.txt
echo "Total sieving time: $sievetime hours" >> summary.txt
echo "Total relation processing time: $relproctime hours" >> summary.txt
echo "Total matrix solve time: $mattime hours" >> summary.txt
echo "Time per square root: $sqrttime hours" >> summary.txt
echo "Total time: $totaltime hours" >> summary.txt

cat summary.txt
