#!/bin/bash
############################################################
# This is a special-case test directory, different from
# the others. It just uses the classical siever. I basically
# keep this test case around in case I want to quickly factor
# a small number to check for trivial breaks in the code and
# to help with debugging. But for a better example of how to
# use GGNFS, look at some of the other directories.
############################################################
NAME=tst150
RFBSIZE=3000
RSS=3000
AFBSIZE=3000
ASS=3000
MPR=19
MPA=19
A0=-250000
A1=250000
BSTEPSIZE=200
RLAMBDA=1.7
ALAMBDA=2.0
NUMLPR=1
NUMLPA=1
KNOWNDIV=1
CACHE=524288
SAVEPAIRS=0
CLEANUP=0
# This is optional - if, for some reason, you happen to
# already know a large prime divisor of the discriminant
# of the polynomial, you can set it here to save getdeps the
# trouble of trying to find it.
#DD="DISC_P_DIV=1693893976025287897489923949"
#####################################
# The path where the binaries are:
GGNFS_BIN_PATH=../../src
#####################################
# Just in case, for some reason, you want to start somewhere else?
STARTB=1


##########################################################
# Probably nothing below here should need to be changed. #
##########################################################
SIEVE=$GGNFS_BIN_PATH/sieve
GETDEPS=$GGNFS_BIN_PATH/getdeps
MATSOLVE=$GGNFS_BIN_PATH/matsolve
SQRT=$GGNFS_BIN_PATH/sqrt
PLOT=$GGNFS_BIN_PATH/autogplot.sh
DEPFILE=deps
DEPINF=depinf
PROCRELS=rels.bin
TMPJOBNAME="tmp.job"
LOGFILE="ggnfs.log"

###########################################################
function plotLP() {
  echo "0, 0" > .lprels
  echo "0, 0" > .rels
  echo "0, 0" > .fulls
  grep "largePrimes:" $LOGFILE > .tmp.out
  numLines=`wc .tmp.out | awk '{ print $1}'`
  for name in `seq $numLines`
  do
    read line
    line=${line#*largePrimes:}
    lps=`echo $line | awk -F' ' '{print $1}'`
    rels=`echo $line | awk -F': ' '{print $2}'`
     diff=`echo "$lps - $rels" | bc`
     echo "$rels, $diff" >> .lprels
  done <.tmp.out
  XAXIS="Total relations"
  YAXIS=""
  export XAXIS
  export YAXIS
  $PLOT relations.jpg "ExcessLargePrimes" .lprels 
#  grep "rels:" $LOGFILE > .tmp.out
#  numLines=`wc .tmp.out | awk '{ print $1}'`
#  for name in `seq $numLines`
#  do
#    read line
#    line=${line#*rels:}
#    rels=`echo $line | awk -F', ' '{print $1}'`
#    fulls=`echo $line | awk -F':' '{print $3}'`
#    echo "$rels, $fulls" >> .fulls
#  done <.tmp.out
#  $PLOT relations.jpg LargePrimes .lprels Relations .rels Fulls .fulls
  rm -f .lprels .rels .fulls .tmp.out
}
######################################################

# We won't take the liberty of deleting a dependency file,
# in case this script is run by accident.
if test -e $DEPFILE ; then
  echo "File $DEPFILE already exists. Delete it and re-run for a fresh factorization"
  exit
fi
if test ! -e $NAME.poly ; then
  echo "Error: Polynomial file $F does not exist!"
  exit
fi
typeset -i RESUME=0
if test -e $PROCRELS.0 ; then
  echo "It appears that an earlier attempt was interrupted in progress. Resume? (y/n)"
  read RESPONSE
  if test $RESPONSE = "Y" ; then RESUME=1; fi
  if test $RESPONSE = "y" ; then RESUME=1; fi;
fi;

if test $RESUME -ne 1 ; then
  # Clean up any junk leftover from an earlier attempt.
  rm -f $LOGFILE 
  rm -f cols* deps* factor.easy lpindex* rels*  spairs.out $NAME.fb spairs.save.gz
  echo $DD > $LOGFILE


  # Prime the loop with the range of b-values we're going to try first.
  typeset -i B0=$STARTB
else
  # Get the B0 value from tmp.job and just restart from there.
  B0LINE=`grep b0 $TMPJOBNAME`;
  typeset -i B0=${B0LINE:4}
  echo "Found $B0LINE in $TMPJOBNAME - resuming from b=$B0"
fi;
typeset -i B1=$B0+$BSTEPSIZE-1

while test ! -e $DEPINF ; do
  # First, create the job file.
  rm -f $TMPJOBNAME
  echo "type: classical" > $TMPJOBNAME
  echo "a0: $A0" >> $TMPJOBNAME
  echo "a1: $A1" >> $TMPJOBNAME
  echo "b0: $B0" >> $TMPJOBNAME
  echo "b1: $B1" >> $TMPJOBNAME
  echo "rlambda: $RLAMBDA" >> $TMPJOBNAME
  echo "alambda: $ALAMBDA" >> $TMPJOBNAME
  echo "cacheSize: $CACHE" >> $TMPJOBNAME
  echo "rfb_size: $RFBSIZE" >> $TMPJOBNAME
  echo "afb_size: $AFBSIZE" >> $TMPJOBNAME
  echo "rss: $RSS" >> $TMPJOBNAME
  echo "ass: $ASS" >> $TMPJOBNAME
  echo "mpr: $MPR" >> $TMPJOBNAME
  echo "mpa: $MPA" >> $TMPJOBNAME
  echo "fbname: $NAME.fb" >> $TMPJOBNAME
  echo "maxLP: $NUMLPR" >> $TMPJOBNAME
  echo "maxLPA: $NUMLPA" >> $TMPJOBNAME
  echo "START_POLY" >> $TMPJOBNAME
  cat $NAME.poly >> $TMPJOBNAME
  echo "END_POLY" >> $TMPJOBNAME
  $SIEVE -fb $NAME.fb -j $TMPJOBNAME
  if test ! -e spairs.out ; then
    echo "Some error ocurred and no relations were found! Examine log file."
    exit
  fi
  $GETDEPS -forceCC -fb $NAME.fb -prel $PROCRELS -newrel spairs.out
  if test $SAVEPAIRS -eq 1 ; then
    gzip -c spairs.out >> spairs.save.gz
  fi
  rm spairs.out
  B0=$B1+1
  B1=$B0+BSTEPSIZE-1
#  $PLOT relations.jpg FinalRels numrels.txt initialFF numFF.txt
  plotLP
done
rm -f $TMPJOBNAME

echo "Doing matrix step..."
$MATSOLVE
if test ! -e $DEPFILE ; then
  echo "Some error occurred and matsolve did not record dependencies."
  exit -1
fi

typeset -i depnum=0
$SQRT $DISC -fb $NAME.fb -deps $DEPFILE -knowndiv $KNOWNDIV -depnum $depnum
while test $? -ne 0 ; do
  depnum=$depnum+1
  echo "Computing square root for dependency $depnum ..."
  $SQRT $DISC -fb $NAME.fb -deps $DEPFILE -knowndiv $KNOWNDIV -depnum $depnum >& /dev/null
done;


# Finally, clean up.
if test $CLEANUP -eq 1 ; then
  rm -f cols* deps* factor.easy lpindex* rels*  spairs.out $NAME.fb
  rm -f numFF.txt numrels.txt depinf
fi

######################################################################
# These techniques were borrowed from:
# http://docsrv.caldera.com:8457/en/AdvBashHowto/io-redirection.html
# among other places.
######################################################################

#######################################################
# Summary data for the factorization
sievetime=0.0 ; grep "SieveTime:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  line=${line#*SieveTime:*}
  sievetime=`echo $sievetime + $line | bc`
done <.tmp.out
sievetime=`echo "scale=5; $sievetime / 3600.0" | bc` # Convert to hours

relproctime=0.0 ; grep "RelProcTime:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  line=${line#*RelProcTime:*}
  relproctime=`echo $relproctime + $line | bc`
done <.tmp.out
relproctime=`echo "scale=5; $relproctime / 3600.0" | bc` # Convert to hours

mattime=0.0 ; grep "BLanczosTime:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  line=${line#*BLanczosTime:*}
  mattime=`echo $mattime + $line | bc`
done <.tmp.out
mattime=`echo "scale=5; $mattime / 3600.0" | bc` # Convert to hours
line=`grep "Initial matrix" $LOGFILE`
initmat=${line#*Initial matrix is*}
line=`grep "Matrix pruned" $LOGFILE`
prunedmat=${line#*Matrix pruned to*}

grep "sqrtTime:" $LOGFILE > .tmp.out
read line <.tmp.out
sqrttime=${line#*sqrtTime:*}
sqrttime=`echo "scale=5; $sqrttime / 3600.0" | bc` # Convert to hours
totaltime=`echo "scale=5; $sievetime+$relproctime+$mattime+$sqrttime" | bc`

grep "rels:" $LOGFILE > .tmp.out
numLines=`wc .tmp.out | awk '{ print $1}'`
for name in `seq $numLines` 
do
  read line
  relsline=${line#*rels:*}
done <.tmp.out
rm -f .tmp.out


echo "Number: $NAME" > summary.txt
echo "Factor base sizes: $RFBSIZE/$AFBSIZE" >> summary.txt
echo "Large prime bits: $MPR/$MPA" >> summary.txt
echo "rff=$RFF, aff=$AFF" >> summary.txt
echo "rss=$RSS, ass=$ASS" >> summary.txt
echo "Sieve region: a=[$A0, $A1], b=[1, $B1]" >> summary.txt
echo "Relations: $relsline" >> summary.txt
echo "Initial matrix obtained: $initmat" >> summary.txt
echo "Pruned matrix: $prunedmat" >> summary.txt
echo "Total sieving time: $sievetime hours" >> summary.txt
echo "Total relation processing time: $relproctime hours" >> summary.txt
echo "Total matrix solve time: $mattime hours" >> summary.txt
echo "Time per square root: $sqrttime hours" >> summary.txt
echo "Total time: $totaltime hours" >> summary.txt

cat summary.txt
